"""
This script fetches weather data from the Met Office API, processes it, and generates an image suitable for epaper display, and might be useful for other
applications as well. It uses the Pillow library for image processing and the requests library for API calls.

See [here](https://datahub.metoffice.gov.uk/docs/f/category/map-images/type/map-images/api-documentation) for more information on the API.

"""

import os
import re
from datetime import datetime, timedelta
from functools import lru_cache
from io import BytesIO
from itertools import groupby
from typing import Dict, List, Optional, Tuple

import requests
from PIL import Image, ImageDraw, ImageFilter

# assert os.getenv('MET_OFFICE_API_KEY') is not None, "MET_OFFICE_API_KEY not set in .env file"
# assert os.getenv('MAP_IMAGES_ORDER_NAME') is not None, "MAP_IMAGES_ORDER_NAME not set in .env file"

BASE_URL = "https://data.hub.api.metoffice.gov.uk/map-images/1.0.0"

### API Client Functions

session = requests.Session()
session.headers.update({"Accept": "application/json", "apikey": f"{os.getenv('MET_OFFICE_API_KEY')}"})


def get_order_latest(order_name: str) -> Dict:
    url = f"{BASE_URL}/orders/{order_name.lower()}/latest"  # pragma: no cover
    # TODO: Network integration testing - requires valid Met Office API key and order
    response = session.get(url)  # pragma: no cover
    if response.status_code == 200:  # pragma: no cover
        return response.json()  # pragma: no cover
    else:  # pragma: no cover
        raise Exception(f"Failed to fetch order status: {response.status_code} {response.text}")  # pragma: no cover


def get_file_meta(order_name: str, file_id: str) -> Dict:
    url = f"{BASE_URL}/orders/{order_name.lower()}/latest/{requests.utils.quote(file_id)}"  # To handle + in the file_id  # pragma: no cover
    # TODO: Network integration testing - requires valid Met Office API key and order
    response = session.get(url)  # pragma: no cover
    if response.status_code == 200:  # pragma: no cover
        return response.json()  # pragma: no cover
    else:  # pragma: no cover
        raise Exception(f"Failed to fetch order status: {response.status_code} {response.text}")  # pragma: no cover


@lru_cache
def get_file(order_name: str, file_id: str) -> bytes:
    url = f"{BASE_URL}/orders/{order_name.lower()}/latest/{requests.utils.quote(file_id)}/data"  # To handle + in the file_id  # pragma: no cover
    # TODO: Network integration testing - requires valid Met Office API key and order
    response = session.get(
        url, headers={**session.headers, **{"Accept": "application/octet-stream"}}
    )  # pragma: no cover
    if response.status_code == 200:  # pragma: no cover
        return response.content  # pragma: no cover
    else:  # pragma: no cover
        raise Exception(f"Failed to fetch order status: {response.status_code} {response.text}")  # pragma: no cover


### Data Filtering

is_my_date = re.compile(r".*_\d{10}$")  # Ends with 10 digits


def filter_relevant_files(order_status: Dict) -> List[Dict]:
    relevant_files = []

    for file in order_status["orderDetails"]["files"]:
        if is_my_date.match(file["fileId"]):
            # print(file['fileId'])
            *parameter_name, time_step, forecast_date = file["fileId"].split("_")

            parameter_name = "_".join(parameter_name)

            time_delta = int(time_step[2:])  # hours
            forecast_year = forecast_date[:4]
            forecast_month = forecast_date[4:6]
            forecast_day = forecast_date[6:8]
            forecast_hour = forecast_date[8:10]
            date = datetime(
                int(forecast_year),
                int(forecast_month),
                int(forecast_day),
                int(forecast_hour),
            )
            date += timedelta(hours=time_delta)

            relevant_files.append(
                {
                    "fileId": file["fileId"],
                    "parameter_name": parameter_name,
                    "date": date,
                    "delta": time_delta,
                }
            )

    # Sort the relevant files by date and delta, so we get the 'freshest' forecast first for a given time
    relevant_files.sort(key=lambda x: (x["date"], x["delta"]))

    return relevant_files


### Image Generation


def make_borders(data: bytes) -> Image.Image:
    # Convert to grayscale
    img = Image.open(BytesIO(data))
    img = img.point(lambda i: 255 if i else 0)
    # Convert to binary image
    img = img.convert("1")
    # Apply edge detection filter
    img = img.filter(ImageFilter.FIND_EDGES)
    img.info["transparency"] = 0

    return img


def make_isolines(data: bytes) -> Image.Image:
    # Convert to grayscale
    img = Image.open(BytesIO(data)).convert("L")
    # Apply edge detection filter
    img = img.filter(ImageFilter.FIND_EDGES)
    # Convert to binary image
    img = img.convert("1")
    img.info["transparency"] = 0

    return img


def make_precipitation(data: bytes) -> Image.Image:
    # Convert to grayscale
    img = Image.open(BytesIO(data)).convert("L")
    img = img.point(lambda i: 255 - i)
    img.info["transparency"] = 0

    return img


def generate_image(
    order_name: str, block: Dict, bounding_box: Optional[Tuple[int, int, int, int]] = (100, 250, 500, 550)
) -> Image.Image:
    # TODO: Network integration testing - requires valid Met Office API key and order
    border = make_borders(get_file(order_name, block["land_cover"]))  # pragma: no cover
    isoline = make_isolines(get_file(order_name, block["mean_sea_level_pressure"]))  # pragma: no cover
    precipitation = make_precipitation(get_file(order_name, block["total_precipitation_rate"]))  # pragma: no cover

    background = Image.blend(isoline.convert("RGBA"), border.convert("RGBA"), alpha=0.5).convert(
        "L"
    )  # pragma: no cover
    background.info["transparency"] = 0  # pragma: no cover

    img = Image.blend(background.convert("RGBA"), precipitation.convert("RGBA"), alpha=0.5)  # pragma: no cover

    if bounding_box:  # pragma: no cover
        img = img.crop(bounding_box)  # pragma: no cover

    img = img.convert("1")  # pragma: no cover

    draw = ImageDraw.Draw(img)  # pragma: no cover
    draw.text((5, 5), block["date"].isoformat(), fill="white")  # pragma: no cover

    return img  # pragma: no cover


def get_uk_precipitation(order_name: str, bounding_box: Optional[Tuple[int, int, int, int]] = None) -> Image.Image:
    """
    Get the latest UK precipitation forecast from the Met Office API and generate an image suitable for epaper display.
    """
    # TODO: Network integration testing - requires valid Met Office API key and order
    order_status = get_order_latest(order_name)  # pragma: no cover
    relevant_files = filter_relevant_files(order_status)  # pragma: no cover

    ## Project the relevant files into forecast blocks keyed by the forecasted date  # pragma: no cover
    forecast_blocks = {  # pragma: no cover
        dt: {blk["parameter_name"]: blk["fileId"] for blk in block}  # pragma: no cover
        for dt, block in groupby(relevant_files, key=lambda x: x["date"])  # pragma: no cover
    }  # pragma: no cover

    ## Get the forecast block with the key closest to the current datetime  # pragma: no cover
    closest_block = min(forecast_blocks.keys(), key=lambda x: abs(x - datetime.now()))  # pragma: no cover
    block = forecast_blocks[closest_block]  # pragma: no cover
    block["date"] = closest_block  # pragma: no cover

    image = generate_image(order_name, block, bounding_box=bounding_box)  # pragma: no cover

    return image  # pragma: no cover



# coding=utf-8
"""
AWS based Asset handling

Includes S3, Kinesis, SSM, SQS, Lambda self-invocation and Redshift querying helpers
"""

import base64
import csv
import io
import json
import logging
import random
import time
from collections import Counter
from gzip import GzipFile
from typing import Any, AnyStr, Callable, Dict, Generator, Iterator, List, Optional, Sequence, SupportsInt, Union

import boto3
import botocore.config
import botocore.exceptions
import psycopg2.extras

from bolster import chunks

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

###
# Global Session Parent
###
# In theory this means a single auth/pool cycle... in theory..
session: Optional[boto3.Session] = None


def start_session(*args, restart=False, **kwargs) -> boto3.Session:
    global session
    if session is None or restart:
        session = boto3.Session(*args, **kwargs)
    else:
        if args or kwargs:
            raise RuntimeWarning(
                """Calling start session with arguments outside of a restart context;
                                    this might not do what you think it does."""
            )
    return session


###
# S3 Helpers
###
# https://stackoverflow.com/a/44478894
# Path Style Addressing for resolution within VPC (needs VPC-endpoint)


def get_s3_client():
    start_session()
    s3 = session.client(
        "s3",
        session.region_name,
        config=botocore.config.Config(
            s3={"addressing_style": "path"},
            connect_timeout=5,
            retries={"max_attempts": 2},
            max_pool_connections=100,
        ),
    )
    return s3


def put_s3(
    obj: Union[Sequence[Dict], io.StringIO],
    key: str,
    bucket: str,
    keys=None,
    gzip: bool = True,
    client=None,
) -> dict:
    """Take either a list of dicts (and dump them as csv to s3) or a
    StringIO buffer (and dump-as-is to s3)

    Args:
      obj: List of records to be written to CSV (or StringIO for direct upload):
      key: Destination Key
      bucket:  Destination Bucket (Default value = S3_ANALYSIS_STORE)
      keys: List of expected keys, can be used to filter or set the order of key entry in the resultant file
       (Default value = None)
      gzip: Compress the object (Default value = True)

    Returns:

    """
    if client is None:
        client = get_s3_client()

    if isinstance(obj, list):
        buffer = io.StringIO()
        if keys is None:  # Use keys inferred (i.e. no given ordering)
            keys = set([k for d in obj for k in d])
        w = csv.DictWriter(buffer, list(keys), extrasaction="ignore")
        w.writeheader()
        for row in obj:
            w.writerow(row)
    elif isinstance(obj, io.StringIO):
        buffer = obj
    else:
        raise (ValueError(f"Invalid type {type(obj)}"))
    buffer.seek(0)

    if key.endswith(".gz"):
        gzip = True
    if gzip:
        if not key.endswith(".gz"):
            key += ".gz"
        with io.BytesIO() as gz_body:
            with GzipFile(None, "wb", 9, gz_body) as gz:
                gz.write(buffer.read().encode("utf-8"))
            gz_body.seek(0)
            return client.upload_fileobj(Bucket=bucket, Key=key, Fileobj=gz_body)

    else:
        return client.put_object(Bucket=bucket, Key=key, Body=buffer.read())


def get_s3(key: str, bucket: str, gzip: bool = True, log_exception=True, client=None) -> io.StringIO:
    """Get Object from S3, generally with gzip decompression included.

    Args:
      key: param bucket:
      gzip: return:
      key: str:
      bucket: str:  (Default value = S3_ANALYSIS_STORE)
      gzip: bool:  (Default value = True)

    Returns:

    """
    if client is None:
        client = get_s3_client()

    try:
        if gzip and not key.endswith(".gz"):
            key += ".gz"
        elif key.endswith(".gz") and not gzip:
            gzip = True
        obj = client.get_object(Bucket=bucket, Key=key)
        if gzip:
            got_text = GzipFile(None, "r", fileobj=io.BytesIO(obj["Body"].read()))
        else:
            got_text = obj["Body"]
        return io.StringIO(got_text.read().decode("utf-8"))
    except Exception as e:
        if log_exception:
            logger.exception(f"Error getting {key}")
        raise e


def check_s3(key: str, bucket: str, client=None) -> bool:
    """https://www.peterbe.com/plog/fastest-way-to-find-out-if-a-file-exists-in-s3

    Args:
      key: str:
      bucket: str:  (Default value = S3_ANALYSIS_STORE)

    Returns:
    """
    if client is None:
        client = get_s3_client()

    response = client.list_objects_v2(Bucket=bucket, Prefix=key)
    for obj in response.get("Contents", []):
        if obj["Key"] == key:
            return True
    else:
        return False


def get_matching_s3_objects(bucket: AnyStr, prefix="", suffix="", client=None) -> Iterator:
    """
    Generate objects in an S3 bucket.

    https://alexwlchan.net/2018/01/listing-s3-keys-redux/

    :param bucket: Name of the S3 bucket.
    :param prefix: Only fetch objects whose key starts with
        this prefix (optional).
    :param suffix: Only fetch objects whose keys end with
        this suffix (optional).
    """
    if client is None:
        client = get_s3_client()

    kwargs = {"Bucket": bucket}

    # If the prefix is a single string (not a tuple of strings), we can
    # do the filtering directly in the S3 API.
    if isinstance(prefix, str):
        kwargs["Prefix"] = prefix

    while True:
        # The S3 API response is a large blob of metadata.
        # 'Contents' contains information about the listed objects.
        resp = client.list_objects_v2(**kwargs)

        try:
            contents = resp["Contents"]
        except KeyError:
            return

        for obj in contents:
            key = obj["Key"]
            if key.startswith(prefix) and key.endswith(suffix):
                yield obj

        # The S3 API is paginated, returning up to 1000 keys at a time.
        # Pass the continuation token into the next response, until we
        # reach the final page (when this field is missing).
        try:
            kwargs["ContinuationToken"] = resp["NextContinuationToken"]
        except KeyError:
            break


def get_matching_s3_keys(bucket: AnyStr, **kwargs) -> Iterator:
    """
    Generate the keys in an S3 bucket.
    https://alexwlchan.net/2018/01/listing-s3-keys-redux/

    :param bucket: Name of the S3 bucket.
    :param prefix: Only fetch keys that start with this prefix (optional).
    :param suffix: Only fetch keys that end with this suffix (optional).
    """
    for obj in get_matching_s3_objects(bucket, **kwargs):
        yield obj["Key"]


def select_from_csv(bucket, key, fields, client=None) -> List:
    if client is None:
        client = get_s3_client()

    # noinspection SqlInjection
    r = client.select_object_content(
        Bucket=bucket,
        Key=key,
        ExpressionType="SQL",
        RequestProgress={"Enabled": True},
        Expression=f"select {','.join(fields)} from s3object s",
        InputSerialization={"CSV": {"FileHeaderInfo": "Use"}},
        OutputSerialization={"JSON": {}},
    )
    results = []
    for event in r["Payload"]:
        if "Records" in event:
            results.append(event["Records"]["Payload"].decode("utf-8"))
        elif "Progress" in event:
            continue
        elif "Stats" in event:
            stats_details = event["Stats"]["Details"]
            logger.info(stats_details)

    results = "".join(results)[:-1].replace("\n", ",")
    return json.loads("[" + results + "]")


def get_latest_key(prefix: str, bucket: str, key: Optional[Callable] = None, client=None) -> str:
    """Walk a given S3 bucket for the lexicographically highest item in the given bucket (defaults to the analysis store
    defined in utils.env)

    Accepts a `key` callable that can be used to decide how the candidate keys are sorted.

    For example, to use loose-versioning, distutils.version.LooseVersion can be passed as the `key` argument

    Args:
      prefix: param bucket:
      key: return:
      prefix: str:
      bucket: str:  (Default value = S3_ANALYSIS_STORE)
      key: Optional[Callable]:  (Default value = None)

    Returns:

    """
    if client is None:
        client = get_s3_client()

    versions = sorted(
        [v["Key"] for v in client.list_objects_v2(Bucket=bucket, Prefix=prefix)["Contents"]],
        key=key,
        reverse=True,
    )
    latest = versions[0]
    return latest


###
# Queueing / Notification (SQD/SNS) Helpers
###
def get_sqs_client():
    start_session()
    sqs = session.client(
        "sqs",
        endpoint_url=f"https://sqs.{session.region_name}.amazonaws.com",
        config=botocore.config.Config(connect_timeout=2, read_timeout=5, retries={"max_attempts": 2}),
    )
    return sqs


def send_to_sqs(records: Iterator, queue: str, chunksize: int = 1, client=None) -> None:
    """Send `records` in chunks of `chunksize` for a given sqs queue in json-serialised format

    Args:
      records: param queue:
      chunksize: return:
      records: Iterator:
      queue: str:
      chunksize: int:  (Default value = 1)

    Returns:

    """
    if client is None:
        client = get_sqs_client()

    n, m = 0, 0
    sqs_incidents_url = client.get_queue_url(QueueName=queue)["QueueUrl"]
    for entry in chunks(records, chunksize):
        client.send_message(QueueUrl=sqs_incidents_url, MessageBody=json.dumps(entry))
        n += len(entry)
        m += 1
    logger.info(f"Delivered {n} items to {queue} in {m} batches")


###
# Shared Secret Manager Helpers
###
_ssm_params = {}


def get_ssm_client():
    start_session()
    ssm_client = session.client(
        "ssm",
        config=botocore.config.Config(
            connect_timeout=2,
            read_timeout=5,
            retries={"max_attempts": 0},
            max_pool_connections=100,
        ),
    )
    return ssm_client


def get_ssm_param(param_name: str, client=None) -> str:
    """Locally memoized getter for configuration parameters stored in the AWS "Simple Systems Manager" (now just
    systems manager) Parameter Store

    Args:
      param_name: return:
      param_name: str:

    Returns:

    """
    global _ssm_params
    client = get_ssm_client()
    if param_name not in _ssm_params:
        param = client.get_parameter(Name=param_name, WithDecryption=True)

        _ssm_params[param_name] = param["Parameter"]["Value"]

    value = _ssm_params[param_name]

    return value


###
# Kinesis/Firehose Helpers
###
def fh_json_decode(content: AnyStr) -> Iterator[Union[Dict, List]]:
    """Customised JSON Decoder for consuming Firehose batched records;

    Firehose doesn't include entry separators between entries, so we intercept the raw_decoder
    on JSONDecodeError and 'skip' over the 'where is my comma?' issue and continue to parse the
    rest of the content until we reach the end of the given content string.

    Args:
      content: AnyStr:

    Returns:

    >>> list(fh_json_decode('{"test":"value"}{"test":"othervalue"}'))
    [{'test': 'value'}, {'test': 'othervalue'}]
    """
    decoder = json.JSONDecoder()
    content_length = len(content)
    decode_index = 0

    while decode_index < content_length:
        try:
            obj, decode_index = decoder.raw_decode(content, decode_index)
            yield obj
        except json.JSONDecodeError:
            # Scan forward and keep trying to decode
            decode_index += 1


def decapsulate_kinesis_payloads(event: Dict) -> List[Dict]:
    """Decapsulate base64 encoded kinesis data records to a list

    Args:
      event: Dict:

    Returns:

    """
    records = []
    for record in event["Records"]:
        try:
            b64payload = base64.b64decode(record["kinesis"]["data"]).decode("utf-8")
            records.append(json.loads(b64payload))
        except Exception:
            logger.exception(f"FAILED {record}")
    return records


def iterate_kinesis_payloads(event: Dict) -> Generator[Dict, None, None]:
    """Iterate over a base64 encoded kinesis data record, yielding entries

    Args:
      event: return:
      event: Dict:

    Returns:

    """
    for record in event["Records"]:
        try:
            b64payload = base64.b64decode(record["kinesis"]["data"]).decode("utf-8")
            yield json.loads(b64payload)
        except Exception:
            logger.exception(f"FAILED {record}")


class KinesisLoader(object):
    """Kinesis batchwise insertion handler with chunking and retry"""

    def __init__(self, batch_size: int = 500, maximum_records: int = None, stream: str = None):
        """
        The default batch_size here is to match the maximum allowed by Kinesis in a PutRecords request
        """
        start_session()
        self.batch_size = min(batch_size, 500)
        self.maximum_records = maximum_records
        self.kinesis_client = session.client(
            "kinesis",
            config=botocore.config.Config(
                connect_timeout=5,
                max_pool_connections=100,
                read_timeout=5,
                retries={"max_attempts": 3},
            ),
        )
        if stream.startswith("arn"):
            stream = stream.split("/")[-1]
        self.stream = stream

    def generate_and_submit(self, items: Iterator, partition_key: str = None) -> SupportsInt:
        """Submit batches of items to the configured stream

        Args:
          items: param partition_key:
          items: Iterator:
          partition_key: str:  (Default value = None)

        Returns:

        """
        counter = 0
        # Simple cutoff here - guaranteed to not send in more than maximum_records, with single batch granularity
        for i, batched_items in enumerate(chunks(items, self.batch_size)):
            records_batch = [
                {
                    "Data": json.dumps(item).encode("utf-8"),
                    "PartitionKey": str(random.random()) if partition_key is None else partition_key,
                }
                for item in batched_items
            ]
            request = {"Records": records_batch, "StreamName": self.stream}

            response = self.kinesis_client.put_records(**request)
            self.submit_batch_until_successful(records_batch, response)

            counter += len(records_batch)
            if counter > 1:
                logger.info("Batch inserted. Total records: {}".format(str(counter)))

        return counter

    def submit_batch_until_successful(self, this_batch: List, response: Dict):
        """If needed, retry a batch of records, backing off exponentially until it goes through

        Args:
          this_batch: List:
          response: Dict:

        Returns:

        """
        retry_interval = 0.25

        failed_record_count = response["FailedRecordCount"]
        while failed_record_count:
            time.sleep(retry_interval)

            # Failed records don't contain the original contents - we have to correlate with the input by position
            failed_records = [this_batch[i] for i, record in enumerate(response["Records"]) if "ErrorCode" in record]

            logger.info(
                "Incrementing exponential back off and retrying {} failed records".format(str(len(failed_records)))
            )
            retry_interval = min(retry_interval * 2, 4)
            request = {"Records": failed_records, "StreamName": self.stream}
            result = self.kinesis_client.put_records(**request)
            failed_record_count = result["FailedRecordCount"]


def send_to_kinesis(records: Iterator[Sequence], stream: str, partition_key: str = None) -> int:
    """Accessory function for the KinesisLoader class

    Args:
      records: Iterator[Sequence]:
      stream: str:
      partition_key: str:  (Default value = None)

    Returns:

    """
    loader = KinesisLoader(stream=stream)
    return loader.generate_and_submit((record for record in records), partition_key=partition_key)


def get_sns_client():
    start_session()
    sns = session.client(
        "sns",
        session.region_name,
        config=botocore.config.Config(connect_timeout=5, retries={"max_attempts": 2}, max_pool_connections=100),
    )
    return sns


def invoke_self_async(event: Dict, context: Any):
    """Have the Lambda invoke itself asynchronously, passing the same event it received originally,
    and tagging the event as 'async' so it's actually processed

    THIS DOES NOT WORK FROM WITHIN A VPC! (There is no lambda-invoke endpoint accessible without poking
    lots of holes in the VPC.

    Args:
      event: Dict:
      context: Any:

    Returns:

    """
    start_session()
    event["async"] = True
    session.client("lambda").invoke(
        FunctionName=context.invoked_function_arn,
        InvocationType="Event",
        Payload=bytes(json.dumps(event).encode("utf-8")),
    )


def query(q: str, redshift_conn_dict: dict, named_cursor="bolster_query_cursor", **kwargs) -> Iterator[Dict]:
    """Helper for making queries to redshift (or any postgres compatible backend)

    .. code-block:: json

        {
          "user":"USERNAME",
          "host":"HOSTNAME",
          "connect_timeout":3,
          "dbname":"DATABASE",
          "port":5439,
          "password":"SUPERSECRETPASSWORD1111"
        }


    This function implements the 'is_local' check if it is getting it's configuration
    dictionary from the parameter store, and will overwrite the 'host' in the store
    with a resolvable hostname for the ALDS datastore.

    Basically, if you're not working on ALDS, in a few very specific locations, or
    are outside the ALDS VPC, give this a sensible dictionary.

    kwargs are passed through as `vars` to the SQL execution, i.e. to be used with
    substitution queries, eg:

    .. code-block:: python

        query("select * from table where id = %(my_id)s", my_id = 14228)


    NOTE! If you use % wildcards (i.e. LIKE '%string'), you're gonna have a bad time... (Use the POSIX regex instead:
    https://docs.aws.amazon.com/redshift/latest/dg/pattern-matching-conditions-posix.html)

    Args:
      q: param redshift_conn_dict:
      kwargs: return:
      q: str:
      redshift_conn_dict: dict:  (Default value = None)
      **kwargs:

    Returns:

    """
    try:
        with psycopg2.connect(**redshift_conn_dict) as conn:
            with conn.cursor(cursor_factory=psycopg2.extras.RealDictCursor, name=named_cursor) as cur:
                cur.execute(q, vars=kwargs if kwargs is not None else {})
                yield from cur
    except BaseException:  # todo workout what the likely exceptions being thrown by postgreql are likely to be
        logger.exception(f"Failed with connection: {redshift_conn_dict}")
        raise


def SQSWrapper(  # noqa: C901
    event,
    context,
    queuename,
    function,
    timeout=60000,
    reinvokelimit=10,
    maxmessages=1,
    raise_exceptions=True,
    deduplicate=False,
    fkwargs={},
    client=None,
):
    if client is None:
        client = get_sqs_client()
    try:
        md5_map = Counter()
        meta_counter = Counter()
        try:
            sqs_url = client.get_queue_url(QueueName=queuename)["QueueUrl"]
        except BaseException:
            logger.exception(f"Error connecting to {queuename}")
            raise
        n = 0
        while context.get_remaining_time_in_millis() > timeout:
            sqs_item = client.receive_message(QueueUrl=sqs_url, MaxNumberOfMessages=maxmessages, WaitTimeSeconds=3)
            if not len(sqs_item.get("Messages", [])):
                if n == 0:
                    logger.debug(f"No messages in {sqs_item}")
                return  # EXIT PATH

            for message in sqs_item["Messages"]:
                receipt_handle = message["ReceiptHandle"]

                if not deduplicate or message["MD5OfBody"] not in md5_map:
                    try:
                        results = function(message["Body"], **fkwargs)
                    except BaseException:
                        results = None
                        logger.exception(f"{function.__name__}: Failed on {message}")
                        if raise_exceptions:
                            raise

                    if isinstance(results, dict):
                        logger.info(f"Got {results}")
                        meta_counter.update(results)
                    elif results is None:
                        pass  # Assume 'None' means function 'failed'
                    else:
                        logger.info(f"Not sure what to do with type {type(results)}:{results}, ignoring")
                else:
                    logger.info(
                        f"Skipping duplicate message: {message['Body']} seen {md5_map[message['MD5OfBody']]} "
                        f"times already"
                    )

                md5_map[message["MD5OfBody"]] += 1

                # Delete received message from queue
                client.delete_message(QueueUrl=sqs_url, ReceiptHandle=receipt_handle)
                n += 1
        logger.info(f"Processed {n} items before wrapping up")
        approx = (
            client.get_queue_attributes(QueueUrl=sqs_url, AttributeNames=["ApproximateNumberOfMessages"])
            .get("Attributes", {})
            .get("ApproximateNumberOfMessages", -1)
        )
        approx = int(approx)
        if approx > reinvokelimit:
            logger.info(f"There are {approx} messages left on the queue; reinvoking asynchronously")
            invoke_self_async(event, context)

        return meta_counter

    except BaseException:
        logger.exception(f"Unhandled Exception in {context.function_name}")
        raise



import datetime

import dateparser
import numpy as np
import pandas as pd
import requests


def get_ni_executive_basic_table() -> pd.DataFrame:
    """
    Takes Data from https://en.wikipedia.org/wiki/Northern_Ireland_Executive#Composition_since_devolution
    Table should be called "Historical composition of the Northern Ireland Executive "

    EG
            Established	Dissolved	Duration	Interregnum
    Executive
    1st	1998-07-01	2002-10-14	1566 days	1667 days
    2nd	2007-05-08	2011-03-24	1416 days	53 days
    3rd	2011-05-16	2016-05-16	1827 days	10 days
    4th	2016-05-26	2017-01-16	235 days	1090 days
    5th	2020-01-11	2022-02-03	754 days	730 days
    6th	2024-02-03	NaT	127 days	NaT

    >>> get_ni_executive_basic_table().dtypes
    Established     datetime64[ns]
    Dissolved       datetime64[ns]
    Duration       timedelta64[ns]
    Interregnum    timedelta64[ns]
    dtype: object

    """
    # Use a custom user agent to avoid Wikipedia 403 errors
    # Wikipedia blocks default pandas/urllib user agents
    headers = {
        "User-Agent": "Bolster Data Science Library/0.3.4 (https://github.com/andrewbolster/bolster; andrew.bolster@gmail.com)"
    }
    url = "https://en.wikipedia.org/wiki/Northern_Ireland_Executive"
    response = requests.get(url, headers=headers)
    response.raise_for_status()
    tables = pd.read_html(response.text)
    tables[4].columns = range(len(tables[4].columns))

    # Get rid of the nasty multi index
    executive_events = tables[4][[0, 1, 2, 3, 4, 5, 6]]
    executive_events.columns = [
        "Executive",
        "Date",
        "Event",
        "vFM",
        "FM",
        "vDFM",
        "DFM",
    ]

    # Get rid of the comments row at the bottom
    executive_events = executive_events[:-1]

    # Clean up the 'Executive' as for some reason wikipedians count the caretakers differently.
    executive_events["Executive"] = executive_events["Executive"].map(lambda x: x.split("(")[0])

    # Use the OFMDFM posts as a proxy for 'active' to flatten out the range of reasons for failure.
    executive_events["Active"] = (
        executive_events[["vFM", "FM", "vDFM", "DFM"]].replace("Vacant", None).replace(np.nan, None).any(axis=1)
    )

    executive_durations = executive_events.groupby(["Executive", "Active"])["Date"].first().unstack()
    executive_durations.columns = ["Dissolved", "Established"]
    executive_durations = executive_durations[reversed(executive_durations.columns)]
    executive_durations = executive_durations.applymap(lambda s: dateparser.parse(s) if isinstance(s, str) else s)
    executive_durations["Duration"] = executive_durations.diff(axis=1).iloc[:, -1:]

    executive_dissolutions = pd.concat(
        [
            executive_durations["Dissolved"],
            executive_durations["Established"].shift(-1),
        ],
        axis=1,
    )
    executive_dissolutions = executive_dissolutions.apply(
        lambda r: r.Established - r.Dissolved
        if not pd.isnull(r.Established)
        else datetime.datetime.today().replace(hour=0, minute=0, second=0, microsecond=0) - r.Dissolved,
        axis=1,
    )

    executive_durations["Interregnum"] = executive_dissolutions

    # Fix last / most recent
    executive_durations.loc[executive_durations.index[-1], "Duration"] = (
        datetime.datetime.today().replace(hour=0, minute=0, second=0, microsecond=0)
        - executive_durations["Established"].iloc[-1]
    )

    return executive_durations



"""Console script for bolster."""

import os
import sys
from datetime import date

import click
import pandas as pd

from . import __version__
from .data_sources.cineworld import get_cinema_listings
from .data_sources.companies_house import get_companies_house_records_that_might_be_in_farset, query_basic_company_data
from .data_sources.eoni import get_results as get_ni_election_results
from .data_sources.metoffice import get_uk_precipitation
from .data_sources.ni_house_price_index import build as get_ni_house_prices
from .data_sources.ni_water import get_postcode_to_water_supply_zone, get_water_quality_by_zone
from .data_sources.wikipedia import get_ni_executive_basic_table


@click.group()
@click.version_option(version=__version__, prog_name="bolster")
@click.option("--verbose", "-v", is_flag=True, help="Enable verbose output")
def cli(verbose, args=None):
    """
    Bolster - A comprehensive Python utility library for Northern Ireland and UK data sources.

    \b
    📊 DATA SOURCES
    Weather & Environment:
        get-precipitation    UK precipitation maps from Met Office
        water-quality        NI water quality by postcode/zone

    \b
    🏛️  Government & Politics:
        ni-executive         NI Executive historical data
        ni-elections         NI Assembly election results (2016-2022)

    \b
    🏢 Business & Property:
        companies-house      UK Companies House data queries
        ni-house-prices      NI house price index data

    \b
    🎬 Entertainment & Lifestyle:
        cinema-listings      Cineworld movie showtimes

    \b
    🛠️  Utilities:
        list-sources         Show all available data sources
        --version           Show version information
        --help              Show this help message

    \b
    📋 QUICK EXAMPLES
        bolster water-quality BT1 5GS       # Water quality for postcode
        bolster ni-executive                 # NI Executive history
        bolster companies-house farset       # Companies at Farset Labs
        bolster list-sources                 # Show all data sources
        bolster --version                    # Show version information

    \b
    💡 TIP: Use 'bolster <command> --help' for detailed command options
    """
    if verbose:
        click.echo("Verbose mode enabled")
    pass


@cli.command()
@click.option(
    "--bounding-box",
    default=None,
    help="Geographic bounding box as 'min_lon,min_lat,max_lon,max_lat' (e.g., '-10.0,49.0,2.0,61.0' for UK)",
)
@click.option(
    "--order-name",
    default=os.getenv("MAP_IMAGES_ORDER_NAME"),
    help="Met Office API order name for precipitation data (or set MAP_IMAGES_ORDER_NAME env var)",
)
@click.option(
    "--output",
    default="precipitation.png",
    help="Output filename for the precipitation map image (default: precipitation.png)",
)
def get_precipitation(bounding_box, order_name, output):
    """
    Download UK precipitation data from the Met Office and save as an image.

    This command retrieves precipitation map data from the UK Met Office API and saves
    it as a PNG image. Requires a Met Office API key and valid order name.

    Environment Variables:
        MET_OFFICE_API_KEY: Your Met Office API key (required)
        MAP_IMAGES_ORDER_NAME: Default order name for precipitation data (optional)

    Examples:
        # Download precipitation map for entire UK
        bolster get-precipitation --order-name "your-order-name"

        # Download for specific region (Northern Ireland)
        bolster get-precipitation --bounding-box "-8.5,54.0,-5.0,55.5" --order-name "your-order-name"

        # Save to custom filename
        bolster get-precipitation --output "ni_rain.png" --order-name "your-order-name"
    """
    # Check for required Met Office API key
    if os.getenv("MET_OFFICE_API_KEY") is None:
        click.echo("❌ Error: MET_OFFICE_API_KEY environment variable is required")
        click.echo("💡 Get your API key from: https://www.metoffice.gov.uk/services/data/datapoint")
        click.echo("   Then set it with: export MET_OFFICE_API_KEY=your_key_here")
        return

    # Validate bounding box format if provided
    if bounding_box is not None:
        try:
            coords = bounding_box.split(",")
            if len(coords) != 4:
                raise ValueError("Must have exactly 4 coordinates")
            min_lon, min_lat, max_lon, max_lat = map(float, coords)
            bounding_box = (min_lon, min_lat, max_lon, max_lat)
            click.echo(f"📍 Using bounding box: {bounding_box}")
        except ValueError as e:
            click.echo(f"❌ Error: Invalid bounding box format: {e}")
            click.echo("💡 Expected format: 'min_lon,min_lat,max_lon,max_lat'")
            click.echo("   Example: '-8.5,54.0,-5.0,55.5' for Northern Ireland")
            return

    # Check for order name
    if order_name is None:
        order_name = os.getenv("MAP_IMAGES_ORDER_NAME")
        if order_name is None:
            click.echo("❌ Error: Order name required but not provided")
            click.echo("💡 Provide it with --order-name or set MAP_IMAGES_ORDER_NAME environment variable")
            click.echo("   Contact Met Office for your specific order name")
            return

    # TODO: API integration testing - requires valid Met Office credentials
    img = get_uk_precipitation(order_name=order_name, bounding_box=bounding_box)  # pragma: no cover
    img.save(output)  # pragma: no cover
    click.echo(f"Precipitation image saved as '{output}'")


@cli.command()
@click.argument("postcode", required=False)
@click.option("--zone-code", help="Water supply zone code (alternative to postcode lookup)")
@click.option(
    "--format",
    "output_format",
    type=click.Choice(["json", "csv", "table"], case_sensitive=False),
    default="table",
    help="Output format (default: table)",
)
def water_quality(postcode, zone_code, output_format):
    """
    Get water quality information for a Northern Ireland postcode or zone.

    Provides water quality data including hardness classification, chemical parameters,
    and compliance information from Northern Ireland Water.

    Examples:
        bolster water-quality BT1 5GS     # Lookup by postcode
        bolster water-quality --zone-code BALM  # Lookup by zone code
        bolster water-quality BT7 --format json  # JSON output
    """
    # Prompt for postcode if neither postcode nor zone code provided
    if not postcode and not zone_code:
        postcode = click.prompt("📍 Enter a Northern Ireland postcode")

    try:
        if postcode and not zone_code:
            # Validate and normalize postcode format
            postcode = postcode.upper().strip()
            if not postcode:
                click.echo("❌ Error: Empty postcode provided")
                return

            # Look up zone code from postcode
            click.echo("🔍 Looking up water supply zone...")
            zone_mapping = get_postcode_to_water_supply_zone()
            postcode_key = postcode.replace(" ", "")
            zone_code = zone_mapping.get(postcode_key, "UNKNOWN")

            if zone_code == "UNKNOWN":
                click.echo(f"❌ Error: Could not find water supply zone for postcode: {postcode}")
                click.echo("💡 Please check the postcode format (e.g., 'BT1 5GS') or try a different postcode")
                click.echo("   Only Northern Ireland postcodes are supported")
                return

            click.echo(f"✅ Postcode {postcode} maps to water supply zone: {zone_code}")

        # Get water quality data
        click.echo("💧 Retrieving water quality data...")
        quality_data = get_water_quality_by_zone(zone_code, strict=False)

        if quality_data.empty:
            click.echo(f"❌ Error: No water quality data available for zone: {zone_code}")
            click.echo("💡 This may be a temporary issue. Please try again later or contact NI Water")
            return

        click.echo("✅ Water quality data retrieved successfully")

        # Output in requested format
        if output_format == "json":
            click.echo(quality_data.to_json(indent=2))
        elif output_format == "csv":
            click.echo(quality_data.to_csv())
        else:  # table format
            click.echo(f"\n💧 Water Quality Data for Zone: {zone_code}")
            click.echo("=" * 50)
            for param, value in quality_data.items():
                click.echo(f"  {param}: {value}")
            click.echo("\n💡 Use --format json or csv for machine-readable output")

    except ConnectionError:
        click.echo("❌ Error: Could not connect to NI Water services")
        click.echo("💡 Please check your internet connection and try again")
    except TimeoutError:
        click.echo("❌ Error: Request timed out")
        click.echo("💡 NI Water services may be temporarily unavailable. Please try again later")
    except Exception as e:
        click.echo(f"❌ Error retrieving water quality data: {e}")
        click.echo("💡 If this error persists, please report it as a bug")


@cli.command()
@click.option(
    "--format",
    "output_format",
    type=click.Choice(["json", "csv", "table"], case_sensitive=False),
    default="table",
    help="Output format (default: table)",
)
@click.option("--save", help="Save data to file (specify filename)")
def ni_executive(output_format, save):
    """
    Get Northern Ireland Executive composition and dissolution data.

    Retrieves historical data about NI Executive periods including establishment
    dates, dissolution dates, duration, and interregnum periods.

    Examples:
        bolster ni-executive                    # Display as table
        bolster ni-executive --format json     # JSON output
        bolster ni-executive --save executive.csv  # Save to CSV file
    """
    try:
        executive_data = get_ni_executive_basic_table()

        if save:
            if save.endswith(".json"):
                executive_data.to_json(save, indent=2, date_format="iso")
            elif save.endswith(".csv"):
                executive_data.to_csv(save)
            else:
                # Default to CSV if no extension specified
                executive_data.to_csv(save)
            click.echo(f"Executive data saved to: {save}")
            return

        # Output in requested format
        if output_format == "json":
            click.echo(executive_data.to_json(indent=2, date_format="iso"))
        elif output_format == "csv":
            click.echo(executive_data.to_csv())
        else:  # table format
            click.echo("\nNorthern Ireland Executive History")
            click.echo("=" * 60)
            click.echo(executive_data.to_string())

    except Exception as e:
        click.echo(f"Error retrieving NI Executive data: {e}")


@cli.command()
@click.option("--site-code", type=int, default=117, help="Cineworld site code (default: 117 for Belfast)")
@click.option("--date", "screening_date", help="Screening date in YYYY-MM-DD format (default: today)")
@click.option(
    "--format",
    "output_format",
    type=click.Choice(["json", "table"], case_sensitive=False),
    default="table",
    help="Output format (default: table)",
)
def cinema_listings(site_code, screening_date, output_format):
    """
    Get current movie listings from Cineworld cinema.

    Retrieves movie showtimes and information for a specific Cineworld location.
    Default location is Belfast (site code 117).

    Examples:
        bolster cinema-listings                     # Belfast, today
        bolster cinema-listings --site-code 105    # Different location
        bolster cinema-listings --date 2024-03-20  # Specific date
    """
    try:
        # Validate and parse screening date
        if screening_date:
            try:
                screening_date = pd.to_datetime(screening_date).date()
                click.echo(f"🎬 Getting listings for {screening_date}")
            except ValueError:
                click.echo("❌ Error: Invalid date format")
                click.echo("💡 Please use YYYY-MM-DD format (e.g., '2024-03-20')")
                return
        else:
            screening_date = date.today()
            click.echo(f"🎬 Getting today's listings ({screening_date})")

        # Validate site code
        if not isinstance(site_code, int) or site_code <= 0:
            click.echo("❌ Error: Invalid site code")
            click.echo("💡 Site code must be a positive integer (default: 117 for Belfast)")
            return

        click.echo(f"🎭 Connecting to Cineworld site {site_code}...")
        listings = get_cinema_listings(site_code=site_code, screening_date=screening_date)

        if not listings:
            click.echo(f"❌ No movie listings found for site {site_code} on {screening_date}")
            click.echo("💡 Possible causes:")
            click.echo("   - Invalid site code (try default 117 for Belfast)")
            click.echo("   - No screenings scheduled for this date")
            click.echo("   - Cineworld API temporarily unavailable")
            return

        click.echo(f"✅ Found {len(listings)} movie listings")

        # Output in requested format
        if output_format == "json":
            import json

            click.echo(json.dumps(listings, indent=2, default=str))
        else:  # table format
            click.echo(f"\n🎬 Cineworld Listings - Site {site_code} - {screening_date}")
            click.echo("=" * 60)
            for i, movie in enumerate(listings, 1):
                click.echo(f"[{i}] {movie.get('title', 'Unknown Title')}")
                if "showtimes" in movie and movie["showtimes"]:
                    showtimes = ", ".join(movie["showtimes"])
                    click.echo(f"    🕐 Showtimes: {showtimes}")
                else:
                    click.echo("    🕐 No showtimes available")

                if "genre" in movie:
                    click.echo(f"    🎭 Genre: {movie['genre']}")

                click.echo("-" * 40)

            click.echo("\n💡 Use --format json for machine-readable output")

    except ConnectionError:
        click.echo("❌ Error: Could not connect to Cineworld services")
        click.echo("💡 Please check your internet connection and try again")
    except TimeoutError:
        click.echo("❌ Error: Request timed out")
        click.echo("💡 Cineworld services may be temporarily unavailable. Please try again later")
    except Exception as e:
        click.echo(f"❌ Error retrieving cinema listings: {e}")
        click.echo("💡 If this error persists, please report it as a bug")


@cli.command()
@click.option(
    "--format",
    "output_format",
    type=click.Choice(["json", "csv"], case_sensitive=False),
    default="csv",
    help="Output format (default: csv)",
)
@click.option("--save", help="Save data to file (specify filename)")
def ni_house_prices(output_format, save):
    """
    Get Northern Ireland house price index data.

    Downloads and processes the latest NI house price statistics from official sources,
    including price trends by property type, region, and time period.

    Examples:
        bolster ni-house-prices                     # Display as CSV
        bolster ni-house-prices --format json      # JSON output
        bolster ni-house-prices --save prices.csv  # Save to file
    """
    try:
        click.echo("Downloading NI house price data... (this may take a moment)")
        house_data = get_ni_house_prices()

        if not house_data:
            click.echo("No house price data available")
            return

        if save:
            if output_format == "json" or save.endswith(".json"):
                import json

                with open(save, "w") as f:
                    json.dump({k: v.to_dict() for k, v in house_data.items()}, f, indent=2, default=str)
            else:
                # Save all tables as separate CSV files
                for table_name, df in house_data.items():
                    filename = f"{save.rsplit('.', 1)[0]}_{table_name}.csv"
                    df.to_csv(filename)
                    click.echo(f"Saved {table_name} to: {filename}")
            return

        # Output summary information
        click.echo(f"\nNI House Price Data - {len(house_data)} tables available:")
        click.echo("=" * 60)
        for table_name, df in house_data.items():
            click.echo(f"{table_name}: {len(df)} rows, {len(df.columns)} columns")
            if not df.empty:
                click.echo(f"  Sample data: {list(df.columns[:3])}")

        click.echo("\nUse --save option to export full data to files")

    except Exception as e:
        click.echo(f"Error retrieving house price data: {e}")


@cli.command()
@click.argument("query", required=False)
@click.option(
    "--format",
    "output_format",
    type=click.Choice(["json", "csv", "table"], case_sensitive=False),
    default="table",
    help="Output format (default: table)",
)
@click.option("--save", help="Save data to file (specify filename)")
def companies_house(query, output_format, save):
    """
    Query UK Companies House data for company information.

    Search for companies by name or get information about companies that might be
    associated with Farset Labs (Belfast hackerspace).

    Examples:
        bolster companies-house "farset"            # Search for companies with 'farset'
        bolster companies-house farset              # Search without quotes
        bolster companies-house --format json      # JSON output
        bolster companies-house --save companies.csv  # Save results
    """
    try:
        if not query:
            # If no query provided, get Farset Labs related companies
            click.echo("🏢 No query provided, retrieving Farset Labs related companies...")
            companies_data = get_companies_house_records_that_might_be_in_farset()
            query_description = "Farset Labs related companies"
        else:
            # Validate and clean query
            query = query.strip()
            if len(query) < 2:
                click.echo("❌ Error: Query too short")
                click.echo("💡 Please provide at least 2 characters for company search")
                return

            # Query for specific company
            click.echo(f"🔍 Searching Companies House for: '{query}'...")
            companies_data = query_basic_company_data(query)
            query_description = f"Companies matching '{query}'"

        if companies_data is None or companies_data.empty:
            search_term = query or "Farset Labs"
            click.echo(f"❌ No companies found for query: {search_term}")
            click.echo("💡 Suggestions:")
            click.echo("   - Try different search terms")
            click.echo("   - Check spelling and try partial company names")
            click.echo("   - Use 'bolster companies-house' without arguments for Farset Labs companies")
            return

        click.echo(f"✅ Found {len(companies_data)} companies")

        if save:
            if output_format == "json" or save.endswith(".json"):
                companies_data.to_json(save, indent=2, date_format="iso")
            elif output_format == "csv" or save.endswith(".csv"):
                companies_data.to_csv(save, index=False)
            else:
                # Default to CSV if no extension specified
                companies_data.to_csv(save, index=False)
            click.echo(f"Companies data saved to: {save}")
            return

        # Output in requested format
        if output_format == "json":
            click.echo(companies_data.to_json(indent=2, date_format="iso"))
        elif output_format == "csv":
            click.echo(companies_data.to_csv(index=False))
        else:  # table format
            click.echo(f"\n{query_description}")
            click.echo("=" * 60)
            if len(companies_data) > 10:
                click.echo(f"Showing first 10 of {len(companies_data)} companies:")
                click.echo(companies_data.head(10).to_string(index=False))
                click.echo(f"\nUse --save option to export all {len(companies_data)} results")
            else:
                click.echo(companies_data.to_string(index=False))

    except ConnectionError:
        click.echo("❌ Error: Could not connect to Companies House API")
        click.echo("💡 Please check your internet connection and try again")
    except TimeoutError:
        click.echo("❌ Error: Request timed out")
        click.echo("💡 Companies House API may be temporarily unavailable. Please try again later")
    except Exception as e:
        error_msg = str(e).lower()
        if "api key" in error_msg or "authentication" in error_msg:
            click.echo("❌ Error: Companies House API authentication failed")
            click.echo("💡 Please check your API credentials or try again later")
        elif "rate limit" in error_msg:
            click.echo("❌ Error: API rate limit exceeded")
            click.echo("💡 Please wait a few minutes before making another request")
        else:
            click.echo(f"❌ Error querying Companies House data: {e}")
            click.echo("💡 If this error persists, please report it as a bug")


@cli.command()
@click.option(
    "--election-year",
    type=click.Choice(["2016", "2017", "2022", "all"], case_sensitive=False),
    default="all",
    help="Filter by election year (default: all)",
)
@click.option(
    "--format",
    "output_format",
    type=click.Choice(["json", "csv", "table"], case_sensitive=False),
    default="table",
    help="Output format (default: table)",
)
@click.option("--save", help="Save data to file (specify filename)")
def ni_elections(election_year, output_format, save):
    """
    Get Northern Ireland Assembly election results (2016-2022).

    Retrieve detailed election results including candidates, parties, constituencies,
    and vote counts for NI Assembly elections.

    Examples:
        bolster ni-elections                        # All election results
        bolster ni-elections --election-year 2022  # 2022 results only
        bolster ni-elections --format json         # JSON output
        bolster ni-elections --save elections.csv  # Save to file
    """
    try:
        click.echo("🗳️ Retrieving NI Assembly election results...")

        # Get election results (the function returns data for all available years)
        election_data = get_ni_election_results()

        if not election_data:
            click.echo("❌ No election data available")
            click.echo("💡 This may be a temporary issue. Please try again later")
            return

        # Filter by year if specified (and not 'all')
        if election_year != "all":
            # This would need to be implemented based on the structure of election_data
            # For now, we'll just note that filtering is requested
            click.echo(f"📊 Filtering results for year: {election_year}")

        click.echo("✅ Election data retrieved successfully")

        # Handle file saving
        if save:
            try:
                if output_format == "json" or save.endswith(".json"):
                    if isinstance(election_data, dict):
                        import json

                        with open(save, "w") as f:
                            json.dump(election_data, f, indent=2, default=str)
                    elif hasattr(election_data, "to_json"):
                        election_data.to_json(save, indent=2, date_format="iso")
                    else:
                        click.echo("❌ Error: Cannot convert election data to JSON format")
                        click.echo("💡 Try using CSV format instead")
                        return
                else:
                    if hasattr(election_data, "to_csv"):
                        election_data.to_csv(save, index=False)
                    else:
                        # If it's a dict of DataFrames, save each separately
                        if isinstance(election_data, dict):
                            for key, df in election_data.items():
                                if hasattr(df, "to_csv"):
                                    filename = f"{save.rsplit('.', 1)[0]}_{key}.csv"
                                    df.to_csv(filename, index=False)
                                    click.echo(f"💾 Saved {key} to: {filename}")
                            return
                click.echo(f"💾 Election data saved to: {save}")
                return
            except PermissionError:
                click.echo(f"❌ Error: Permission denied writing to {save}")
                click.echo("💡 Check file permissions or choose a different location")
                return
            except Exception as e:
                click.echo(f"❌ Error saving file: {e}")
                click.echo("💡 Try a different filename or location")
                return

        # Output in requested format
        if output_format == "json":
            import json

            if isinstance(election_data, dict):
                click.echo(json.dumps(election_data, indent=2, default=str))
            elif hasattr(election_data, "to_json"):
                click.echo(election_data.to_json(indent=2, date_format="iso"))
            else:
                click.echo("Warning: Cannot display election data in JSON format")
        elif output_format == "csv":
            if hasattr(election_data, "to_csv"):
                click.echo(election_data.to_csv(index=False))
            else:
                click.echo("Warning: Cannot display election data in CSV format")
        else:  # table format
            click.echo("\nNI Assembly Election Results")
            if election_year != "all":
                click.echo(f"Year: {election_year}")
            click.echo("=" * 60)

            if hasattr(election_data, "to_string"):
                # If it's a DataFrame
                if len(election_data) > 20:
                    click.echo(f"Showing first 20 of {len(election_data)} records:")
                    click.echo(election_data.head(20).to_string(index=False))
                    click.echo(f"\nUse --save option to export all {len(election_data)} results")
                else:
                    click.echo(election_data.to_string(index=False))
            elif isinstance(election_data, dict):
                # If it's a dict of DataFrames or other data
                for key, value in election_data.items():
                    click.echo(f"\n{key}:")
                    click.echo("-" * 40)
                    if hasattr(value, "to_string"):
                        if len(value) > 10:
                            click.echo(f"First 10 of {len(value)} records:")
                            click.echo(value.head(10).to_string(index=False))
                        else:
                            click.echo(value.to_string(index=False))
                    else:
                        click.echo(str(value))
            else:
                click.echo(str(election_data))

    except Exception as e:
        click.echo(f"Error retrieving election data: {e}")


@cli.command()
def list_sources():
    """
    List all available data sources and their descriptions.

    Shows a comprehensive overview of all data sources available in the Bolster library,
    organized by category with brief descriptions of what data each source provides.
    """
    click.echo("\nBolster - Available Data Sources")
    click.echo("=" * 50)

    click.echo("\n📊 WEATHER & ENVIRONMENT")
    click.echo("  get-precipitation    UK precipitation maps from Met Office API")
    click.echo("                       Requires MET_OFFICE_API_KEY environment variable")

    click.echo("\n💧 WATER & UTILITIES")
    click.echo("  water-quality        NI water quality data by postcode or zone")
    click.echo("                       Chemical parameters, hardness, compliance info")

    click.echo("\n🏛️  GOVERNMENT & POLITICS")
    click.echo("  ni-executive         NI Executive composition and dissolution history")
    click.echo("                       Establishment dates, duration, interregnum periods")
    click.echo("  ni-elections         NI Assembly election results (2016-2022)")
    click.echo("                       Candidates, parties, constituencies, vote counts")

    click.echo("\n🏢 BUSINESS & PROPERTY")
    click.echo("  companies-house      UK Companies House company data queries")
    click.echo("                       Company search, Farset Labs related companies")
    click.echo("  ni-house-prices      NI house price index data from official sources")
    click.echo("                       Price trends by property type, region, time period")

    click.echo("\n🎬 ENTERTAINMENT & LIFESTYLE")
    click.echo("  cinema-listings      Cineworld movie listings and showtimes")
    click.echo("                       Default: Belfast (site 117), supports other locations")

    click.echo("\n🔧 DATA SOURCE MODULES")
    click.echo("  bolster.data_sources.metoffice         - UK Met Office API integration")
    click.echo("  bolster.data_sources.ni_water          - NI Water quality data")
    click.echo("  bolster.data_sources.wikipedia         - NI Executive Wikipedia scraping")
    click.echo("  bolster.data_sources.ni_house_price_index - NI house price statistics")
    click.echo("  bolster.data_sources.cineworld         - Cineworld cinema API")
    click.echo("  bolster.data_sources.eoni              - Electoral Office NI data")
    click.echo("  bolster.data_sources.companies_house   - UK Companies House API")

    click.echo("\n💡 USAGE EXAMPLES")
    click.echo("  bolster water-quality BT1 5GS          # Water quality by postcode")
    click.echo("  bolster ni-executive --format json     # Executive data as JSON")
    click.echo("  bolster companies-house farset          # Search for Farset companies")
    click.echo("  bolster ni-elections --election-year 2022  # 2022 election results")
    click.echo("  bolster cinema-listings --date 2024-03-20  # Movie listings for date")
    click.echo("  bolster --help                          # General help")
    click.echo("  bolster <command> --help               # Command-specific help")

    click.echo(f"\nBolster v{__version__} - Northern Ireland & UK Data Sources")


if __name__ == "__main__":
    sys.exit(cli())  # pragma: no cover



"""
Working with Northern Ireland Water Quality Data

Updated to use modern OpenDataNI data sources after the original
NIWater API endpoints were deprecated.

Current data sources:
- Water quality results: Static CSV data from OpenDataNI
- Zone mapping: Legacy postcode to zone lookup (still available)

"""

import csv
import logging
from typing import Dict, Optional
from urllib.error import HTTPError

import pandas as pd
import requests

from bolster import backoff

# Legacy postcode to zone mapping (still functional)
POSTCODE_DATASET_URL = "https://admin.opendatani.gov.uk/dataset/38a9a8f1-9346-41a2-8e5f-944d87d9caf2/resource/f2bc12c1-4277-4db5-8bd3-b7bb027cc401/download/postcode-v-zone-lookup-by-year.csv"

# Modern water quality data from OpenDataNI
WATER_QUALITY_CSV_URL = "https://admin.opendatani.gov.uk/dataset/38a9a8f1-9346-41a2-8e5f-944d87d9caf2/resource/02d85526-c082-482c-b205-a318f97fd18d/download/2024-ni-water-customer-tap-supply-point-results.csv"

T_HARDNESS = pd.CategoricalDtype(["Soft", "Moderately Soft", "Slightly Hard", "Moderately Hard"], ordered=True)

INVALID_ZONE_IDENTIFIER = "No Zone Identified"

# Cache for water quality data to avoid repeated downloads
_water_quality_cache: Optional[pd.DataFrame] = None


@backoff((HTTPError, RuntimeError))
def get_water_quality_csv_data() -> pd.DataFrame:
    """
    Get the latest water quality CSV data from OpenDataNI.

    This function downloads and caches the complete water quality dataset
    which contains all results from customer tap supply points.

    Returns:
        pd.DataFrame: Complete water quality data with columns:
            - Year: Sample year
            - Sample Location: Location description
            - Site Code: Unique site identifier
            - Site Name: Human-readable site name
            - Sample Id Text: Unique sample identifier
            - Sample Date: Date of sample collection
            - Postcode: Postcode for the sample location
            - Parameter: Water quality parameter name (e.g., 'Total Hardness (mg/l)')
            - PCV Limit: Prescribed Concentration Value limit
            - Result: Numeric test result
            - Report Value: Formatted result value
            - Units: Unit of measurement

    Raises:
        HTTPError: If the CSV download fails
        RuntimeError: If no data is found
    """
    global _water_quality_cache

    if _water_quality_cache is not None:
        return _water_quality_cache

    logging.info(f"Downloading water quality data from {WATER_QUALITY_CSV_URL}")

    with requests.get(WATER_QUALITY_CSV_URL, stream=True) as r:
        r.raise_for_status()
        _water_quality_cache = pd.read_csv(r.url)

        if _water_quality_cache.empty:
            raise RuntimeError("No water quality data found in CSV")

    logging.info(f"Loaded {len(_water_quality_cache)} water quality records")
    return _water_quality_cache


def _site_code_to_zone_code(site_code: str) -> str:
    """
    Convert OpenDataNI site code to legacy zone code format.

    The new CSV data uses site codes like 'BALM' while the legacy API
    used zone codes like 'ZS0101'. This function provides a mapping
    between the two systems for backward compatibility.

    Args:
        site_code: Site code from the CSV data

    Returns:
        Legacy-style zone code
    """
    # For now, we'll use the site code directly as the zone identifier
    # This maintains the function signatures but uses the new data
    return site_code


def _create_legacy_format_series(site_data: pd.DataFrame, zone_code: str) -> pd.Series:
    """
    Convert CSV format data to legacy API format for backward compatibility.

    The original API returned a pandas Series with specific keys.
    This function recreates that format using the new CSV data.

    Args:
        site_data: DataFrame containing all parameters for a specific site
        zone_code: The zone/site identifier

    Returns:
        pd.Series in the legacy format
    """
    # Create the legacy format series
    result_data = {}

    # Get the first row for site information
    if not site_data.empty:
        first_row = site_data.iloc[0]
        result_data["Water Supply Zone"] = first_row.get("Site Name", f"Site {zone_code}")

    # Map CSV parameters to legacy format
    parameter_mapping = {
        "Total hardness": "Total Hardness (mg/l)",  # Note: CSV uses "Total hardness", legacy used "Total Hardness (mg/l)"
        "Magnesium": "Magnesium (mg/l)",
        "Potassium": "Potassium (mg/l)",
        "Calcium": "Calcium (mg/l)",
    }

    # Extract parameter values
    for _, row in site_data.iterrows():
        parameter = row.get("Parameter", "")
        if parameter in parameter_mapping:
            legacy_key = parameter_mapping[parameter]
            result_data[legacy_key] = row.get("Report Value", row.get("Result", ""))

    # Calculate derived values if we have the basic measurements
    try:
        if "Total Hardness (mg/l)" in result_data:
            # The CSV data appears to be in mg/l format, which we need to interpret
            # Assuming this is already CaCO3 equivalent (which is standard for water hardness)
            hardness_mg_l = float(result_data["Total Hardness (mg/l)"])
            hardness_caco3 = hardness_mg_l  # Assume it's already in CaCO3 equivalent

            # Add the CaCO3 format that legacy API provided
            result_data["Total Hardness (mg CaCO3/l)"] = str(hardness_caco3)

            # Calculate degrees of hardness
            result_data["Clark English Degrees"] = f"{hardness_caco3 / 14.3:.1f}"
            result_data["French Degrees"] = f"{hardness_caco3 / 10.0:.1f}"
            result_data["German Degrees"] = f"{hardness_caco3 / 17.8:.1f}"

            # Classify hardness based on standard UK classifications
            if hardness_caco3 <= 60:
                hardness_class = "Soft"
                dishwasher_setting = "1"
            elif hardness_caco3 <= 120:
                hardness_class = "Moderately Soft"
                dishwasher_setting = "2"
            elif hardness_caco3 <= 180:
                hardness_class = "Slightly Hard"
                dishwasher_setting = "3"
            else:
                hardness_class = "Moderately Hard"
                dishwasher_setting = "4"

            result_data["NI Hardness Classification"] = hardness_class
            result_data["Dishwasher Setting"] = dishwasher_setting

    except (ValueError, KeyError) as e:
        # If we can't calculate derived values, just skip them
        logging.debug(f"Could not calculate hardness classifications: {e}")
        pass

    return pd.Series(result_data, name=zone_code)


@backoff((HTTPError, RuntimeError))
def get_postcode_to_water_supply_zone() -> Dict[str, str]:
    """
    Using data from OpenDataNI to generate a map from NI Postcodes to Water Supply Zone

    >>> zones = get_postcode_to_water_supply_zone()
    >>> len(zones)
    49006

    Zones is keyed off postcode to a Water Supply Zone
    >>> zones['BT1 1AA']
    'ZS0107'

    There are much fewer zones than postcodes
    >>> len(set(zones.values()))
    65

    And many postcodes that aren't associated with any zone
    >>> len([k for k,v in zones.items() if v == INVALID_ZONE_IDENTIFIER])
    97

    """

    with requests.get(POSTCODE_DATASET_URL, stream=True) as r:
        lines = (line.decode("utf-8") for line in r.iter_lines())
        reader = csv.DictReader(lines)
        keys = reader.fieldnames[:2]  # Take POSTCODE and first year
        zones = dict(([row[k] for k in keys] for row in reader))
        if not zones:
            raise RuntimeError("No data found")

    return zones


def get_water_quality_by_zone(zone_code: str, strict=False) -> pd.Series:
    """
    Get the latest Water Quality for a given Water Supply Zone.

    Now uses modern OpenDataNI CSV data instead of the deprecated HTML API.
    The zone_code can be either a legacy zone code (like 'ZS0101') or a
    site code from the CSV data (like 'BALM').

    Args:
        zone_code: Water supply zone identifier or site code
        strict: If True, raise ValueError for invalid zones

    Returns:
        pd.Series: Water quality data in legacy API format with indices like:
            - Water Supply Zone: Human-readable zone name
            - Total Hardness (mg/l): Hardness as mg/l
            - Total Hardness (mg CaCO3/l): Hardness as mg CaCO3/l
            - Clark English Degrees: English degrees of hardness
            - French Degrees: French degrees of hardness
            - German Degrees: German degrees of hardness
            - NI Hardness Classification: Categorical hardness level
            - Dishwasher Setting: Recommended dishwasher setting

    Raises:
        ValueError: If zone_code is invalid and strict=True
        HTTPError: If CSV data cannot be downloaded
        RuntimeError: If CSV data is empty

    Example usage:
        data = get_water_quality_by_zone('BALM')  # Using CSV site code
        print(data['Water Supply Zone'])
        print(data['NI Hardness Classification'])
    """
    try:
        # Get the full CSV dataset
        water_quality_df = get_water_quality_csv_data()

        # Try to find data for this zone/site code
        # First try as site code, then try as a potential zone mapping
        site_data = water_quality_df[water_quality_df["Site Code"] == zone_code]

        if site_data.empty:
            # Try to find by site name containing the zone code
            site_data = water_quality_df[water_quality_df["Site Name"].str.contains(zone_code, case=False, na=False)]

        if site_data.empty:
            # No data found for this zone
            if strict:
                raise ValueError(f"Potentially invalid Water Supply Zone {zone_code}")
            else:
                logging.warning(f"Potentially invalid Water Supply Zone {zone_code}")
                return pd.Series(name=zone_code)

        # Convert to legacy format
        return _create_legacy_format_series(site_data, zone_code)

    except (HTTPError, RuntimeError) as err:
        # Handle data source errors
        if strict:
            raise ValueError(f"Unable to retrieve data for Water Supply Zone {zone_code}") from err
        else:
            logging.warning(f"Unable to retrieve data for Water Supply Zone {zone_code}: {err}")
            return pd.Series(name=zone_code)


def get_water_quality() -> pd.DataFrame:
    """
    Get a DataFrame of Water Quality Data from OpenDataNI.

    This function now uses the modern CSV data source instead of the deprecated
    HTML API. It returns water quality data for all available sites.

    Returns:
        pd.DataFrame: Water quality data with one row per site/zone.
            Columns include hardness measurements, classifications, and
            other water quality parameters. The 'NI Hardness Classification'
            column uses categorical data type with proper ordering.

    Raises:
        HTTPError: If CSV data cannot be downloaded
        RuntimeError: If CSV data is empty

    Example usage:
        df = get_water_quality()
        print(df.shape)  # Number of rows and columns
        print(df['NI Hardness Classification'].value_counts(sort=False))

        # Show available sites
        print(df.index.tolist())  # Site codes

        # Get hardness data
        hardness_summary = df.groupby('NI Hardness Classification').size()
        print(hardness_summary)
    """
    try:
        # Get the full CSV dataset
        water_quality_df = get_water_quality_csv_data()

        # Get unique site codes (equivalent to the old zone concept)
        unique_sites = water_quality_df["Site Code"].unique()

        logging.info(f"Processing water quality data for {len(unique_sites)} sites")

        # Process each site to create legacy-format series
        site_series_list = []
        for site_code in unique_sites:
            if pd.isna(site_code) or site_code == "":
                continue

            site_data = water_quality_df[water_quality_df["Site Code"] == site_code]
            if not site_data.empty:
                try:
                    series = _create_legacy_format_series(site_data, site_code)
                    if not series.empty:
                        site_series_list.append(series)
                except Exception as e:
                    logging.warning(f"Error processing site {site_code}: {e}")
                    continue

        if not site_series_list:
            raise RuntimeError("No valid water quality data could be processed")

        # Combine all series into a DataFrame
        df = pd.DataFrame(site_series_list)

        # Ensure NI Hardness Classification uses proper categorical type
        if "NI Hardness Classification" in df.columns:
            df = df.astype({"NI Hardness Classification": T_HARDNESS})

        logging.info(f"Created water quality DataFrame with {len(df)} sites and {len(df.columns)} parameters")
        return df

    except Exception as e:
        logging.error(f"Failed to get water quality data: {e}")
        raise



from datetime import date, datetime, timedelta, timezone
from typing import Union


def round_to_week(dt: Union[datetime, date]) -> date:
    """Return a date for the Monday before the given date

    Args:
      dt: return:
      dt: datetime:

    Returns:

    >>> round_to_week(datetime(2018,8,9,12,1))
    datetime.date(2018, 8, 6)
    >>> round_to_week(date(2018,8,9))
    datetime.date(2018, 8, 6)
    """
    return (dt.date() if isinstance(dt, datetime) else dt) - timedelta(days=dt.weekday())


def round_to_month(dt: Union[datetime, date]) -> date:
    """Return a date for the first day of the month of a given date

    Args:
      dt: return:
      dt: datetime:

    Returns:

    >>> round_to_month(datetime(2018,8,9,12,1))
    datetime.date(2018, 8, 1)
    >>> round_to_month(date(2018,8,9))
    datetime.date(2018, 8, 1)
    """
    dt = dt.date() if isinstance(dt, datetime) else dt

    return date(dt.year, dt.month, 1)


def utc_midnight_on(dt: datetime) -> datetime:
    """Some services don't like timezones, so this helper function converts `datetime.date` and
    `datetime.datetime` objects to a `datetime.datetime` object corresponding to UTC Midnight
    on that date.


    Pays primary attention to the actual Date of the input, regardless of if the combination
    of given-time and timezone would roll over into another date.

    Args:
      dt: return:
      dt: datetime:

    >>> utc_midnight_on(datetime(2018,9,1,12,12))
    datetime.datetime(2018, 9, 1, 0, 0, tzinfo=datetime.timezone.utc)

    >>> utc_midnight_on(datetime(2018,9,1,12,12, tzinfo=timezone(timedelta(hours=-13))))
    datetime.datetime(2018, 9, 1, 0, 0, tzinfo=datetime.timezone.utc)
    """
    return datetime.combine(dt.date(), datetime.min.time()).replace(tzinfo=timezone.utc)



"""
Basic statistics and data frame helpers.

Simple functions for common data manipulation tasks:
- add_totals/drop_totals: manage row/column totals in DataFrames
- top_n: truncate DataFrames to top N rows with 'others' aggregation
- fix_datetime_tz_columns: strip timezone info from datetime columns

Plus distribution fitting in the distributions submodule.
"""

from typing import AnyStr

import pandas as pd


def add_totals(
    df: pd.DataFrame,
    column_total: AnyStr = "total",
    row_total: AnyStr = "total",
    inplace=True,
):
    """
    Add Row and Column totals to a dataframe (in place)

    >>> add_totals(pd.DataFrame([[0,1,2],[3,4,5]]))
           0  1  2  total
    0      0  1  2      3
    1      3  4  5     12
    total  3  5  7     15

    >>> add_totals(pd.DataFrame([[0,1,2],[3,4,5]]),'ctot', 'rtot')
          0  1  2  rtot
    0     0  1  2     3
    1     3  4  5    12
    ctot  3  5  7    15


    >>> df = pd.DataFrame([[0,1,2],[3,4,5]])
    >>> add_totals(df, inplace=False)
           0  1  2  total
    0      0  1  2      3
    1      3  4  5     12
    total  3  5  7     15

    >>> df
       0  1  2
    0  0  1  2
    1  3  4  5

    """
    if not inplace:
        df = df.copy(deep=True)
    df.loc[column_total] = df.sum(numeric_only=True, axis=0)
    df.loc[:, row_total] = df.sum(numeric_only=True, axis=1)

    return df


def drop_totals(
    df: pd.DataFrame,
    column_total: AnyStr = "total",
    row_total: AnyStr = "total",
    inplace=True,
) -> pd.DataFrame:
    """
    Remove Row and Column totals from a dataframe (in place)

    Parameters
    ----------
    df : pd.DataFrame
        The DataFrame from which to remove totals.
    column_total : AnyStr, optional
        The name of the column total, by default "total".
    row_total : AnyStr, optional
        The name of the row total, by default "total".
    inplace : bool, optional
        Whether to modify the DataFrame in place, by default True.

    Returns
    -------
    pd.DataFrame
        The DataFrame with totals removed.

    Examples
    --------
    >>> df = pd.DataFrame({'A': [1, 2, 3], 'B': [4, 5, 6], 'total': [5, 7, 9]})
    >>> df.loc['total'] = [6, 15, 21]
    >>> drop_totals(df)
       A  B
    0  1  4
    1  2  5
    2  3  6
    """
    if not inplace:
        df = df.copy(deep=True)

    if column_total in df.columns:
        df = df.drop(columns=[column_total])

    if row_total in df.index:
        df = df.drop(index=[row_total])

    return df


def fix_datetime_tz_columns(df: pd.DataFrame, inplace=True) -> pd.DataFrame:
    """
    Strip Timezone information from relevant datetime columns in a dataframe

    Parameters
    ----------
    df
    inplace (bool)


    Returns
    -------
    df

    """
    if not inplace:
        df = df.copy(deep=True)
    date_columns = df.select_dtypes(include=["datetime64[ns, UTC]", "datetimetz"]).columns
    for date_column in date_columns:
        df[date_column] = df[date_column].dt.tz_localize(None)
    return df


def top_n(df: pd.DataFrame, n: int, others: AnyStr = "others") -> pd.DataFrame:
    """
    Truncate the DataFrame to the top 'n' rows, summing all subsequent rows into an 'others' row.

    Parameters
    ----------
    df : pd.DataFrame
        The DataFrame to truncate.
    n : int
        The number of top rows to keep.

    Returns
    -------
    pd.DataFrame
        The truncated DataFrame with an 'others' row.

    Examples
    --------
    >>> df = pd.DataFrame({'A': [1, 2, 3, 4, 5], 'B': [5, 4, 3, 2, 1]})
    >>> top_n(df, 3) # doctest: +NORMALIZE_WHITESPACE
           A  B
    0      1  5
    1      2  4
    2      3  3
    others 9  3
    """
    if n >= len(df):
        return df

    top_df = df.iloc[:n]
    others_df = df.iloc[n:].sum(numeric_only=True)
    if isinstance(others_df, (pd.Series, pd.DataFrame)):
        others_df.name = others
    else:
        others_df = pd.Series(others_df, name=others)
    return pd.concat([top_df, others_df.to_frame().T])



"""
Utility bits and pieces.

Random helpful functions that don't fit anywhere else:
- timed: decorator to time function execution
- TqdmLoggingHandler: logging that plays nice with tqdm progress bars
- web: resilient web scraping helpers
- dt: datetime utilities
- io: file/data helpers
- aws/azure: cloud platform helpers
"""

import datetime
import logging
import time
from functools import wraps
from typing import Any, Callable, TypeVar

import tqdm

F = TypeVar("F", bound=Callable[..., Any])

version_no = f"{(datetime.date.today() - datetime.date(1988, 5, 17)).total_seconds() / 31557600:.2f}"


class TqdmLoggingHandler(logging.Handler):
    """
    Custom logging handler that uses tqdm to display log messages.
    i.e. `logging.getLogger().addHandler(TqdmLoggingHandler())`
    """

    def __init__(self, level=logging.NOTSET):
        super().__init__(level)

    def emit(self, record):
        try:
            msg = self.format(record)
            tqdm.tqdm.write(msg)
            self.flush()
        except Exception:
            self.handleError(record)


def timed(func: F) -> F:
    """This decorator prints the execution time for the decorated function."""

    @wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        start = time.time()
        logging.info(f"Launching {func.__name__}")
        result = func(*args, **kwargs)
        end = time.time()
        logging.info(f"{func.__name__} ran in {round(end - start, 2)}s")
        return result

    return wrapper



"""
Azure Utils
"""

from urllib.parse import urlparse


def az_file_url_to_query_components(url: str):
    """
    Helper function to parse an Azure file URL into its components to then be used by `pandas`/`dask`/`fsspec` etc.

    >>> az_file_url_to_query_components("https://storageaccount.blob.core.windows.net/container/file_path.parquet")
    {'storage_account': 'storageaccount', 'container': 'container', 'file_path': 'file_path.parquet'}
    """

    p = urlparse(url)
    assert not p.params, f"Invalid Params: {p.params}"
    assert not p.fragment, f"Invalid Fragment: {p.fragment}"
    assert not p.query, f"Invalid Params: {p.query}"

    netlocs = p.netloc.split(".")
    assert len(netlocs) == 5, f"Invalid netlocs: {p.netloc}: Not long enough"
    assert netlocs[2:] == [
        "core",
        "windows",
        "net",
    ], f"Invalid netlocs: {p.netloc} should end in core.windows.net"
    assert netlocs[1] in [
        "blob",
        "dfs",
    ], f"Invalid netlocs: {p.netloc} should be one of blob/dfs"

    storage_account = netlocs[0]
    _, container, *paths = p.path.split("/")  # path starts with a / so p.path.split('/')[0] == ''
    file_path = "/".join(paths)

    return dict(storage_account=storage_account, container=container, file_path=file_path)



import warnings
from time import time

import numpy as np
import pandas as pd
import scipy.stats as st
from tqdm.auto import tqdm


def _get_available_distributions(include_slow=False):
    """Get list of available scipy.stats distributions, filtering out incompatible ones."""
    # Define distributions that should be available in most scipy versions
    distribution_names = [
        "alpha",
        "anglit",
        "arcsine",
        "beta",
        "betaprime",
        "bradford",
        "burr",
        "cauchy",
        "chi",
        "chi2",
        "cosine",
        "dgamma",
        "dweibull",
        "erlang",
        "expon",
        "exponnorm",
        "exponweib",
        "exponpow",
        "f",
        "fatiguelife",
        "fisk",
        "foldcauchy",
        "foldnorm",
        "genlogistic",
        "genpareto",
        "gennorm",
        "genexpon",
        "genextreme",
        "gausshyper",
        "gamma",
        "gengamma",
        "genhalflogistic",
        "gilbrat",
        "gompertz",
        "gumbel_r",
        "gumbel_l",
        "halfcauchy",
        "halflogistic",
        "halfnorm",
        "halfgennorm",
        "hypsecant",
        "invgamma",
        "invgauss",
        "invweibull",
        "johnsonsb",
        "johnsonsu",
        "ksone",
        "kstwobign",
        "laplace",
        "levy",
        "levy_l",
        "logistic",
        "loggamma",
        "loglaplace",
        "lognorm",
        "lomax",
        "maxwell",
        "mielke",
        "nakagami",
        "ncx2",
        "ncf",
        "nct",
        "norm",
        "pareto",
        "pearson3",
        "powerlaw",
        "powerlognorm",
        "powernorm",
        "rdist",
        "reciprocal",
        "rayleigh",
        "rice",
        "recipinvgauss",
        "semicircular",
        "t",
        "triang",
        "truncexpon",
        "truncnorm",
        "tukeylambda",
        "uniform",
        "vonmises",
        "vonmises_line",
        "wald",
        "weibull_min",
        "weibull_max",
        "wrapcauchy",
    ]

    # Filter to only include distributions that actually exist in current scipy version
    available_distributions = []
    for name in distribution_names:
        if hasattr(st, name):
            available_distributions.append(getattr(st, name))

    slow_distributions = []
    if hasattr(st, "levy_stable"):
        slow_distributions.append(st.levy_stable)

    if include_slow:
        available_distributions.extend(slow_distributions)

    return available_distributions


# Create models from data https://stackoverflow.com/questions/6620471/fitting-empirical-distribution-to-theoretical-ones-with-scipy-python
def best_fit_distribution(data, bins=200, ax=None, include_slow=False, discriminator="sse"):
    """Model data by finding best fit distribution to data"""
    # Get histogram of original data
    y, x = np.histogram(data, bins=bins, density=True)
    x = (x + np.roll(x, -1))[:-1] / 2.0

    # Get available distributions for current scipy version
    DISTRIBUTIONS = _get_available_distributions(include_slow)

    # Best holders
    best_distribution = st.norm
    best_params = (0.0, 1.0)
    best_discriminator_value = np.inf

    times = {}

    # Estimate distribution parameters from data
    for distribution in tqdm(DISTRIBUTIONS):
        # Try to fit the distribution
        try:
            # Ignore warnings from data that can't be fit
            with warnings.catch_warnings():
                warnings.filterwarnings("ignore")

                # fit dist to data
                start = time()
                print(distribution.name, end=" ")
                params = distribution.fit(data)
                times[distribution.name] = time() - start
                print(f"took {int(times[distribution.name])}")

                # Separate parts of parameters
                arg = params[:-2]
                loc = params[-2]
                scale = params[-1]

                if discriminator == "sse":
                    # Calculate fitted PDF and error with fit in distribution
                    pdf = distribution.pdf(x, loc=loc, scale=scale, *arg)
                    discriminator_value = np.sum(np.power(y - pdf, 2.0))
                else:
                    raise RuntimeError("You didn't finish this and you were planning on doing KS discrimination next")

                # if axis pass in add to plot
                try:
                    if ax:
                        pd.Series(pdf, x).plot(ax=ax)
                except Exception:
                    pass

                # identify if this distribution is better
                if best_discriminator_value > discriminator_value > 0:
                    best_distribution = distribution
                    best_params = params
                    best_discriminator_value = discriminator_value
                    print(f"New best, {distribution.name} and got an sse of {discriminator_value}")

        except Exception:
            pass

    return (best_distribution.name, best_params)



"""
Data sources - mostly Northern Ireland focused stuff.

Various scrapers and API wrappers for NI government data:
- ni_water: Water quality by postcode/zone
- metoffice: UK precipitation maps
- wikipedia: NI Executive history parsing
- ni_house_price_index: Housing market data
- cineworld: Movie listings
- companies_house: UK company data
- eoni: Electoral data

Most have corresponding CLI commands.
"""



"""
Working with the Norther Ireland House Price index data

Original Source: https://www.nisra.gov.uk/statistics/housing-community-and-regeneration/northern-ireland-house-price-index

See [here](https://andrewbolster.info/2022/03/NI-House-Price-Index.html) for more details

Generic problems fixed;

* Tables offset from header with notes and annotations
* Inconsistent offset
* Inconsistent treatment of quarterly periods (Q1 2020/Quarter 1 2020/ (2020,Q1) etc)
* Inconsistent header alignment (i.e. leaving sub-category annotations on what is actually the period index header)
"""

import logging
import re
from functools import partial
from typing import Dict, Text
from urllib.parse import urlparse

import bs4
import pandas as pd
import requests

DEFAULT_URL = "https://www.finance-ni.gov.uk/publications/ni-house-price-index-statistical-reports"
TABLE_TRANSFORMATION_MAP = {}


def get_source_url(base_url=DEFAULT_URL) -> Text:
    base_content = requests.get(base_url).content
    base_soup = bs4.BeautifulSoup(base_content, features="html.parser")
    source_url = None
    for a in base_soup.find_all("a"):
        if a.attrs.get("href", "").lower().endswith("xlsx"):
            source_url = a.attrs["href"]
            if source_url.startswith("/"):  # Relative URL
                source_url = urlparse(base_url)._replace(path=source_url).geturl()
            return source_url
    if source_url is None:
        raise RuntimeError(f"Could not find valid/relevant Excel source file on {base_url}")


def pull_sources(base_url=DEFAULT_URL) -> Dict[Text, pd.DataFrame]:
    """
    Pull raw NI House Price Index Excel from finance-ni.gov.uk listing

    Parameters
    ----------
    base_url

    Returns
    -------

    """
    source_url = get_source_url(base_url)
    if source_url is not None:
        ni_house_price_index = pd.read_excel(source_url, sheet_name=None)  # Load all worksheets in
    else:
        raise RuntimeError(f"Could not find valid/relevant Excel source file on {base_url}")

    return ni_house_price_index


def basic_cleanup(df: pd.DataFrame, offset=1) -> pd.DataFrame:
    """
    Generic cleanup operations for NI HPI data;
    * Re-header from Offset row and translate table to eliminate incorrect headers
    * remove any columns with 'Nan' or 'None' in the given offset-row
    * If 'NI' appears and all the values are 100, remove it.
    * Remove any rows below and including the first 'all nan' row (gets most tail-notes)
    * If 'Sale Year','Sale Quarter' appear in the columns, replace with 'Year','Quarter' respectively
    * For Year; forward fill any none/nan values
    * If Year/Quarter appear, add  a new composite 'Period' column with a PeriodIndex columns representing the
        year/quarter (i.e. 2022-Q1)
    * Reset and drop the index
    * Attempt to infer the new/current column object types

    Parameters
    ----------
    df
    offset

    Returns
    -------

    """
    df = df.copy()
    # Re-header from row 1 (which was row 3 in excel)
    new_header = df.iloc[offset]
    df = df.iloc[offset + 1 :]
    df.columns = new_header

    # remove 'NaN' trailing columns
    df = df[df.columns[pd.notna(df.columns)]]

    # 'NI' is a usually hidden column that appears to be a checksum;
    # if it's all there and all 100, remove it, otherwise, complain.
    # (Note, need to change this 'if' logic to just 'if there's a
    # column with all 100's, but cross that bridge later)
    if "NI" in df:
        assert df["NI"].all() and df["NI"].mean() == 100, "Not all values in df['NI'] == 100"
        df = df.drop("NI", axis=1)

    # Strip rows below the first all-nan row, if there is one
    # (Otherwise this truncates the tables as there is no
    # idxmax in the table of all 'false's)
    if any(df.isna().all(axis=1)):
        idx_first_bad_row = df.isna().all(axis=1).idxmax()
        df = df.loc[: idx_first_bad_row - 1]

    # By Inspection, other tables use 'Sale Year' and 'Sale Quarter'
    if set(df.keys()).issuperset({"Sale Year", "Sale Quarter"}):
        df = df.rename(columns={"Sale Year": "Year", "Sale Quarter": "Quarter"})

    # For 'Year','Quarter' indexed pages, there is an implied Year
    # in Q2/4, so fill it downwards
    if set(df.keys()).issuperset({"Year", "Quarter"}):
        df["Year"] = df["Year"].astype(float).ffill().astype(int)
        df = df.dropna(how="any", subset=["Year", "Quarter"])

        # In Pandas we can represent Y/Q combinations as proper datetimes
        # https://stackoverflow.com/questions/53898482/clean-way-to-convert-quarterly-periods-to-datetime-in-pandas
        df.insert(
            loc=0,
            column="Period",
            value=pd.PeriodIndex(df.apply(lambda r: f"{r.Year}-{r.Quarter}", axis=1), freq="Q"),
        )

    # reset index, try to fix dtypes, etc, (this should be the last
    # operation before returning!
    df = df.reset_index(drop=True).infer_objects()

    return df


def cleanup_contents(df: pd.DataFrame) -> pd.DataFrame:
    """
    Fix Contents table of NI HPI Stats
    * Shift/rebuild headers
    * Strip Figures because they're gonna be broken anyway

    Parameters
    ----------
    df

    Returns
    -------

    """
    new_header = df.iloc[0]
    df = df[1:].copy()
    df.columns = [*new_header[:-1], "Title"]
    # df['Worksheet Name'] = df['Worksheet Name'].str.replace('Figure', 'Fig')
    df = df[df["Worksheet Name"].str.startswith("Table")]

    return df


TABLE_TRANSFORMATION_MAP["Contents"] = cleanup_contents
# NI HPI Trends
TABLE_TRANSFORMATION_MAP["Table 1"] = basic_cleanup


def cleanup_price_by_property_type_agg(df: pd.DataFrame) -> pd.DataFrame:
    """
    NI HPI & Standardised Price Statistics by Property Type (Aggregate Table)

    Standard cleanup with a split to remove trailing index date data

    Parameters
    ----------
    df

    Returns
    -------

    """
    df = basic_cleanup(df)
    df.columns = [c.split("\n")[0] for c in df.columns]
    return df


# NI HPI & Standardised Price Statistics by Property Type
TABLE_TRANSFORMATION_MAP["Table 2"] = cleanup_price_by_property_type_agg


def cleanup_price_by_property_type(df: pd.DataFrame) -> pd.DataFrame:
    """
    NI HPI & Standardised Price Statistics by Property Type (Per Class)

    Standard cleanup, removing the property class from the table columns

    Parameters
    ----------
    df

    Returns
    -------

    """
    df = basic_cleanup(df)
    new_columns = []
    for c in df.columns:
        if c.endswith("Price Index"):
            new_columns.append("Index")
        elif c.endswith("Standardised Price"):
            new_columns.append("Price")
        else:
            new_columns.append(c)

    df.columns = new_columns

    return df


# NI {property type} Property Price Index
TABLE_TRANSFORMATION_MAP[re.compile("Table 2[a-z]")] = cleanup_price_by_property_type

# NI HPI & Standardised Price Statistics by New/Existing Resold Dwelling Type
TABLE_TRANSFORMATION_MAP["Table 3"] = cleanup_price_by_property_type_agg
TABLE_TRANSFORMATION_MAP[
    re.compile("Table 3[a-z]")  # Table 3c Overridden below
] = cleanup_price_by_property_type_agg


def cleanup_with_munged_quarters_and_total_rows(df: pd.DataFrame, offset=3) -> pd.DataFrame:
    """
    Number of Verified Residential Property Sales

    * Regex 'Quarter X' to 'QX' in future 'Sales Quarter' column
    * Drop Year Total rows
    * Clear any Newlines from the future 'Sales Year' column
    * call `basic_cleanup` with offset=3

    Parameters
    ----------
    df

    Returns
    -------

    """
    df = df.copy()
    df.iloc[:, 1] = df.iloc[:, 1].str.replace("Quarter ([1-4])", r"Q\1", regex=True)
    df = df[~df.iloc[:, 1].str.contains("Total", na=False)]
    # Lose the year new-lines (needs astype because non str lines are
    # correctly inferred to be ints, so .str methods nan-out
    with pd.option_context("mode.chained_assignment", None):
        df.iloc[:, 0] = df.iloc[:, 0].astype(str).str.replace("\n", "")

    df = basic_cleanup(df, offset=offset)
    return df


TABLE_TRANSFORMATION_MAP["Table 3c"] = partial(cleanup_with_munged_quarters_and_total_rows, offset=4)

# Table 4: Number of Verified Residential Property Sales
TABLE_TRANSFORMATION_MAP["Table 4"] = cleanup_with_munged_quarters_and_total_rows


def cleanup_with_LGDs(df: pd.DataFrame) -> pd.DataFrame:
    """
    Standardised House Price & Index for each Local Government District Northern Ireland
    * Build multi-index of LGD / Metric [Index,Price]

    """
    # Basic Cleanup first
    df = basic_cleanup(df)
    # Build multi-index of LGD / Metric [Index,Price]
    # Two inner-columns per LGD
    lgds = (
        df.columns[3:]
        .str.replace(" Standardised HPI", " HPI")
        .str.replace(" HPI", "")
        .str.replace(" Standardised Price", "")
        .unique()
    )
    df.columns = [
        *df.columns[:3],
        *pd.MultiIndex.from_product([lgds, ["Index", "Price"]], names=["LGD", "Metric"]),
    ]
    return df


# Table 5: Standardised House Price & Index for each Local Government District Northern Ireland
TABLE_TRANSFORMATION_MAP["Table 5"] = cleanup_with_LGDs


def cleanup_merged_year_quarters_and_totals(df: pd.DataFrame) -> pd.DataFrame:
    """
    Table 5a: Number of Verified Residential Property Sales by Local Government District
    * Remove 'Total' rows (i.e. Yearly totals)
    * Replace 'Quarter N' with 'QN' in the Sale Quarter column
    """
    # Safety first
    df = df.copy()

    # remove rows containing 'Total' in the column 1:
    df = df[~df.iloc[:, 1].str.contains("Total", na=False)]

    # refactor 'Quarter N' to 'QN'
    df.iloc[:, 1] = df.iloc[:, 1].str.replace("Quarter ", "Q", regex=False)

    # Standard cleanup
    df = basic_cleanup(df, offset=2)

    return df


# Table 5a: Number of Verified Residential Property Sales by Local Government District
TABLE_TRANSFORMATION_MAP["Table 5a"] = cleanup_merged_year_quarters_and_totals
# Table 6: Standardised House Price & Index for all Urban and Rural areas in NI
TABLE_TRANSFORMATION_MAP["Table 6"] = basic_cleanup


def cleanup_missing_year_quarter(df: pd.DataFrame) -> pd.DataFrame:
    """
    Table 7: Standardised House Price & Index for Rural Areas of Northern Ireland by drive times
    * Insert Year/Quarter future-headers
    * Clean normally
    # TODO THIS MIGHT BE VALID FOR MULTIINDEXING ON DRIVETIME/[Index/Price]
    """
    df = df.copy()
    df.iloc[1, 0] = "Year"
    df.iloc[1, 1] = "Quarter"
    df = basic_cleanup(df)
    return df


# Table 7: Standardised House Price & Index for Rural Areas of Northern Ireland by drive times
TABLE_TRANSFORMATION_MAP["Table 7"] = cleanup_missing_year_quarter

# Table 8: Number of Verified Residential Property Sales of properties in urban and rural areas and properties in rural areas by drive times witihn towns of 10,000 or more and within 1 hour of Belfast
TABLE_TRANSFORMATION_MAP["Table 8"] = cleanup_merged_year_quarters_and_totals

# Table 9: NI Average Sales Prices
TABLE_TRANSFORMATION_MAP["Table 9"] = basic_cleanup
TABLE_TRANSFORMATION_MAP[re.compile("Table 9[a-z]")] = cleanup_missing_year_quarter

# Table 10x: Number of Verified Residential Property Sales by Type in {LDG}
TABLE_TRANSFORMATION_MAP[re.compile("Table 10[a-z]")] = cleanup_merged_year_quarters_and_totals


def transform_sources(source_df: Dict[Text, pd.DataFrame]) -> Dict[Text, pd.DataFrame]:
    """
    Cleanup all the tables from the NI Housing Price Index, conforming to the best attempt at a 'standard'

    Parameters
    ----------
    source_df

    Returns
    -------

    """
    dest_df = {}
    for table_key, table_transformer in TABLE_TRANSFORMATION_MAP.items():
        try:
            table = None  # Catch looping variables in debug
            if isinstance(table_key, re.Pattern):
                for table in source_df:
                    if table_key.match(table):
                        dest_df[table] = table_transformer(source_df[table])
            else:
                dest_df[table_key] = table_transformer(source_df[table_key])
            logging.info(f"Transformed {table_key} with {table_transformer}")
        except Exception as e:
            raise RuntimeError(f"Error transforming {table_key} / {table_transformer}") from e

    return dest_df


def build() -> Dict[Text, pd.DataFrame]:
    """
    Pulls and Cleans up the latest Northern Ireland House Price Index Data

    Returns
    -------

    """
    source_dfs = pull_sources()
    dest_dfs = transform_sources(source_dfs)
    return dest_dfs



"""
Working with Election Office NI Data

The bits that we can; this module is primarily concerned with the ingestion of NI Assembly election results from 2003
onwards (where possible in a vaguely reliable automated way)

Hitlist:
[X] 2022
[X] 2017
[X] 2016
[ ] 2011
[ ] 2007
[ ] 2003

"""

import datetime
import re
from typing import AnyStr, Dict, Iterable, Optional, Union

import pandas as pd
import requests
from bs4 import BeautifulSoup

from bolster.utils.web import get_excel_dataframe, ua

#
_headers = {
    "user-agent": f"User-Agent: {ua} Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, "
    f"like Gecko) Chrome/91.0.4472.114 Safari/537.36"
}
_base_url = "https://www.eoni.org.uk"


def get_page(path: AnyStr) -> BeautifulSoup:
    """
    For a given path (within EONI.org.uk), get the response as a BeautifulSoup instance

    Note:
        EONI is trying to block people from scraping and will return a 403 error if you don't pass a 'conventional' user agent

    >>> page = get_page("/Elections/")
    >>> page.find('title').contents[0].strip()
    'Elections | The Electoral Office for Northern Ireland'

    """
    res = requests.get(_base_url + path, headers=_headers)
    res.raise_for_status()
    page = BeautifulSoup(res.content, features="html.parser")
    return page


def find_xls_links_in_page(page: BeautifulSoup) -> Iterable[AnyStr]:
    """
    Walk through a BeautifulSoup page and iterate through '(XLS)' suffixed links

    (Primarily Used for 'Results' pages within given elections)

    #WTF Was starting to do some consistency checks between elections to make sure all is kosher, and was wondering why I had a Strangford listing in 2017 but not 2022;
    # As a cross-check on the result page, I walk the links in the right colum of the page, looking for links that have text that ends (XLS). Pretty simple you might think. Except the Strangford link ends in (XLS  and then a random closing ) text string is added to the end.

    >>> page = get_page("/results-data/ni-assembly-election-2022-results/")
    >>> len(list(find_xls_links_in_page(page)))
    18
    >>> next(find_xls_links_in_page(page))
    'https://www.eoni.org.uk/media/omtlpqow/ni-assembly-election-2022-result-sheet-belfast-east-xls.xlsx'

    """
    for _p in page.select_one(".c-article--main").find_all("a", href=True):
        if "xls" in _p.contents[0].lower():
            yield _base_url + _p.attrs["href"]


def normalise_constituencies(cons_str: str) -> str:
    """
    Some constituencies change names or cases etc;

    Use this function to take external/unconventional inputs and project them into a normalised format

    >>> normalise_constituencies('Newry & Armagh')
    'newry and armagh'

    """
    return cons_str.lower().replace(" & ", " and ")


def get_metadata_from_df(
    df: pd.DataFrame,
) -> Dict[str, Union[int, str, datetime.datetime]]:
    """
    Extract Ballot metadata from the table header(s) of an XLS formatted result sheet, as output from `get_excel_dataframe`

    # TODO this could probably be done better as a `dataclass`

    Returns:
        dict of
            'stage': int,
            'date': datetime
            'constituency': str (lower)
            'eligible_electorate': int
            'votes_polled': int
            'number_to_be_elected': int
            'invalid_votes': int
            'electoral_quota': int
    """

    stage_n_catcher = re.compile(r"^Stage (\d+)")

    metadata = {
        "stage": int(re.match(stage_n_catcher, df.columns[5]).group(1)),
        # should have been just int(df.columns[5].split()[-1])., but someone insisted on messing up 2017
        "date": df.columns[10],
        "constituency": normalise_constituencies(df.iloc[0, 3]),
        "eligible_electorate": int(df.iloc[1, 3]),
        "votes_polled": int(df.iloc[2, 3]),
        "number_to_be_elected": int(df.iloc[1, 6]),
        "total_valid_votes": int(df.iloc[2, 6]),
        "invalid_votes": int(df.iloc[1, 9]),
        "electoral_quota": int(df.iloc[1, 12]),
    }
    return metadata


def get_candidates_from_df(df: pd.DataFrame) -> pd.DataFrame:
    """
    Extract Candidates name and party columns from first stage sheet
    """
    candidates_df = df.iloc[9:29, 2:4]
    candidates_df.columns = ["candidate_name", "candidate_party"]
    return candidates_df.replace(0, None).dropna().reset_index(drop=True)


def get_stage_votes_from_df(df: pd.DataFrame) -> pd.DataFrame:
    """
    Extract the votes from each stage as a mapped column for each stage, i.e. stages 1...N
    """
    stages = get_metadata_from_df(df)["stage"]
    stage_df = (
        pd.concat({n: extract_stage_n_votes(df, n) for n in range(stages)})
        .unstack()
        .T.replace(0, None)
        .dropna(how="all")
    )
    return stage_df


def get_stage_transfers_from_df(df: pd.DataFrame) -> pd.DataFrame:
    """
    Extract the transfers from each stage as a mapped column for each stage, i.e. stages 2...N
    """
    stages = get_metadata_from_df(df)["stage"]
    stage_df = (
        pd.concat({n: extract_stage_n_transfers(df, n) for n in range(stages)})
        .unstack()
        .T.replace(0, None)
        .dropna(how="all")
    )
    return stage_df


def extract_stage_n_votes(df: pd.DataFrame, n: int) -> Optional[pd.Series]:
    """
    Extract the votes from a given stage N

    Note: This will include trailing, unaligned `Nones` which must be cleaned up at the Ballot level
    """
    if n == 0:
        return None
    if n < 10:
        row_offset = 9
        col_offset = 4 + (2 * (n - 1))
    else:
        row_offset = 55
        col_offset = 6 + (2 * (n - 10))

    return df.iloc[row_offset : row_offset + 20, col_offset].reset_index(drop=True)


def extract_stage_n_transfers(df: pd.DataFrame, n: int) -> Optional[pd.Series]:
    """
    Extract the votes from a given stage N

    Note: This will include trailing, unaligned `Nones` which must be cleaned up at the Ballot level
    Stage Transfers are associated with the 'next' stage, i.e. stage 1 has no transfers
    """
    if n <= 1:
        return None
    if n < 10:
        row_offset = 9
        col_offset = 5 + (2 * (n - 2))
    else:
        row_offset = 55
        col_offset = 5 + (2 * (n - 10))

    return df.iloc[row_offset : row_offset + 20, col_offset].reset_index(drop=True)


def get_results_from_sheet(sheet_url: AnyStr) -> Dict[str, Union[pd.DataFrame, dict]]:
    df = get_excel_dataframe(sheet_url, requests_kwargs={"headers": _headers})
    metadata = get_metadata_from_df(df)
    candidates = get_candidates_from_df(df)
    stage_votes = get_stage_votes_from_df(df)
    stage_transfers = get_stage_transfers_from_df(df)

    return {
        "candidates": candidates,
        "stage_votes": stage_votes,
        "stage_transfers": stage_transfers,
        "metadata": metadata,
    }


def get_results(year: int) -> Dict[str, Union[pd.DataFrame, dict]]:
    results_listing_dir = "/results-data/"
    results_listing_path = {
        2022: "ni-assembly-election-2022-results/",
    }

    results = {}
    results_listing_page = get_page(results_listing_dir + results_listing_path[year])
    for sheet_url in find_xls_links_in_page(results_listing_page):
        data = get_results_from_sheet(sheet_url)
        results[data["metadata"]["constituency"]] = data
    return results



import io
import logging
import zipfile
from io import BytesIO
from typing import Dict, Generator, Optional, Tuple

import pandas as pd
import requests
from waybackpy import WaybackMachineCDXServerAPI, exceptions

from . import version_no

ua = f"@Bolster/{version_no} (+http://bolster.online/)"

session = requests.Session()
session.headers.update({"User-Agent": ua})


def get_last_valid(url: str) -> str:
    return WaybackMachineCDXServerAPI(url).oldest().archive_url


def resilient_get(url: str, **kwargs) -> requests.Response:
    """
    Attempt a get, but if it fails, try using the wayback machine to get the last valid version and get that.
    If all else fails, raise a HTTPError from the inner "NoCDXRecordFound" exception
    """

    try:
        res = requests.get(url, **kwargs)
        res.raise_for_status()
    except requests.HTTPError as outer_err:
        try:
            last_valid = get_last_valid(url)
        except exceptions.NoCDXRecordFound as inner_err:
            raise outer_err from inner_err
        res = requests.get(last_valid, **kwargs)
        res.raise_for_status()
        logging.warning(f"Failed to get {url} directly, successfully used waybackmachine to get {last_valid}")
    return res


def get_excel_dataframe(
    file_url: str, requests_kwargs: Optional[Dict] = None, read_kwargs: Optional[Dict] = None
) -> pd.DataFrame:
    if requests_kwargs is None:
        requests_kwargs = {}
    if read_kwargs is None:
        read_kwargs = {}

    with requests.get(file_url, **requests_kwargs) as response:
        response.raise_for_status()
        data = BytesIO(response.content)
        df = pd.read_excel(data, **read_kwargs)
        return df


def download_extract_zip(url: str) -> Generator[Tuple[str, io.BufferedReader], None, None]:
    """
    Download a ZIP file and extract its contents in memory
    yields (filename, file-like object) pairs
    """
    with session.get(url, stream=True) as response:
        response.raise_for_status()
        with zipfile.ZipFile(io.BytesIO(response.content)) as thezip:
            for zipinfo in thezip.infolist():
                with thezip.open(zipinfo) as thefile:
                    yield zipinfo.filename, thefile



from typing import AnyStr, BinaryIO, Dict, Union

import pandas as pd


def save_xls(dict_df: Dict[AnyStr, pd.DataFrame], path: Union[str, BinaryIO]) -> None:
    """
    Save a dictionary of dataframes to an excel file, with each dataframe as a separate page
    """

    with pd.ExcelWriter(path) as writer:
        for key in dict_df:
            dict_df[key].to_excel(writer, key)



"""
This module provides functions to retrieve cinema listings from the Cineworld API.

The main function in this module is `get_cinema_listings`, which takes a site code and a screening date as input and returns a dictionary containing the cinema listings for that date.

Site Code 117 maps to Belfast, you're on your own for the rest.

Example usage:
    cinema_listings = get_cinema_listings(117)
    list(cinema_listings[0].keys())
    # Output: ['id', 'name', 'length', 'posterLink', 'videoLink', 'link', 'weight', 'releaseYear', 'attributeIds', 'date', 'site_code']

"""

from datetime import date
from typing import Any, Dict, List

import requests

from ..utils.web import session


def get_cinema_listings(site_code: int = 117, screening_date: date = date.today()) -> List[Dict[str, Any]]:
    """
    Get cinema listings from the Cineworld API.

    Args:
        site_code (int): The site code of the cinema. Defaults to 117; Belfast
        screening_date (date): The date for which to retrieve the listings. Defaults to today's date.

    Returns:
        dict: A dictionary containing the cinema listings.

    Raises:
        requests.exceptions.RequestException: If there was an error making the API request.

    >>> cinema_listings = get_cinema_listings(117)
    >>> set(cinema_listings[0].keys()).issuperset({'id', 'name','link','weight','releaseYear','releaseDate','attributeIds','date','site_code'})
    True
    >>> list(cinema_listings[0].keys()) # This is likely to break from upstream changes
    ['id', 'name', 'length', 'posterLink', 'videoLink', 'link', 'weight', 'releaseYear', 'releaseDate', 'attributeIds', 'date', 'site_code']

    """
    if screening_date is None:
        screening_date = date.today()

    if not isinstance(screening_date, date):
        try:
            screening_date = date.fromisoformat(screening_date)
        except ValueError as e:
            raise ValueError("screening_date must be a date object or a string in the format 'YYYY-MM-DD'") from e

    url = f"https://www.cineworld.co.uk/uk/data-api-service/v1/quickbook/10108/film-events/in-cinema/{site_code}/at-date/{screening_date}"

    try:
        response = session.get(url)
        response.raise_for_status()
        listings = response.json()["body"]["films"]
        for list in listings:
            list["date"] = screening_date
            list["site_code"] = site_code
        return listings
    except requests.exceptions.RequestException as e:
        raise e
